# -*- coding: utf-8 -*-
"""
@author: Peter Morgan <pete@daffodil.uk.com>
"""

## This simplifies the import...
## and will allows for pyside or alike later

from PyQt5 import QtCore
from PyQt5.QtCore import Qt

from PyQt5 import QtGui
from PyQt5 import QtWidgets
from PyQt5 import QtChart
from PyQt5 import QtNetwork
from PyQt5 import QtWebEngineWidgets
#from PyQt5 import QtSql


from PyQt5.QtCore import pyqtSignal, pyqtSlot




# -*- coding: utf-8 -*-




version = "0.1-alpha"

##==============================================
## Below  vars are initialsed in MainWindow.__init__

args = None
"""Command args from :ref:`ogt-desktop.py` """

mainWin = None
"""Pointer to the :class:`~ogtgui.main_window.MainWindow`"""

settings = None
"""Global :class:`settings.Settings` instance"""

server = None
"""Pointer to http client"""

ags = None
# """Pointer to an :class:`ogtgui.ags_models.AgsObject` instance"""



# -*- coding: utf-8 -*-

import os
import glob
import glob
import json
import yaml

from ogt import EXAMPLES_DIR, USER_TEMP


def to_yaml(data):
    """Serializes python data to a :ref:`yaml` text string ascii atmo

    :type data: dict or list
    :param data: the python data to be encoded
    :return: a `tuple` containing

             - `str` with the encoded json
             - An `Error` message if encoding error, otherwise `None`
    """

    try:
        yaml_str = yaml.safe_dump(data, default_flow_style=False)
        return yaml_str, None

    except Exception as e:
        return None, "Error: %s" % str(e)


def write_yaml_file(file_path, data):
    """Saves python data in a :ref:`json` encoded file

    :param file_path: The relative or absolute path of file to save.
    :type file_path: str
    :param data: The python data to save
    :type data: dict or list
    :return: `Error` message if write error, otherwise `None`
    """
    with open(file_path, "w") as f:
        yaml_str, err = to_yaml(data)
        if err:
            return err
        f.write(yaml_str)
        f.close()
    return None


def to_json(data, minify=False):
    """Serializes python data to a :ref:`json` string

    :type data: dict or list
    :param data: the python data to be encoded
    :type minify: bool
    :param minify:

            - When **`False`** the json string is minimized with no spaces, new lines etc.
            - When **`True`** the json string is human readable indented with four spaces, and sorted by key.

            .. note:: **Important**

                - By default this project uses **`minify=False`**.
                - For versioning (eg git), it is recommended to use **`minify=False`** as the string
                  will always be the same, ie sorted keys, and indentation


    :return: a `tuple` containing

             - `str` with the encoded json
             - An `Error` message if encoding error, otherwise `None`
    """
    # todo: catch json error
    if minify:
        return json.dumps(data, ensure_ascii=True, separators=(',', ':')), None
    return json.dumps(data, ensure_ascii=True, indent=4, sort_keys=True, separators=(',', ': ')), None


def write_json_file(file_path, data, minify=False):
    """Saves python data in a :ref:`json` encoded file

    :param file_path: The relative or absolute path of file to save.
    :type file_path: str
    :param data: The python data to save
    :type data: dict or list
    :param minify: see :func:`~ogt.utils.to_json`
    :type minify: bool
    :return: `Error` message if write error, otherwise `None`
    """
    with open(file_path, "w") as f:
        json_str, err = to_json(data, minify=minify)
        if err:
            return err
        f.write(json_str)
        f.close()
    return None


def read_json_file(file_path):
    """Read and decodes a :ref:`json` encoded file


    :param file_path: The relative or absolute path of file to read.
    :type file_path: str
    :rtype: tuple
    :return:
             - A decoded python `dict` or `list`
             - An `Error` message if file or decoding error, otherwise `None`
    """
    try:
        with open(file_path, "r") as f:
            data = json.load(f)
            return data, None
    except Exception as e:
        return None, str(e)


def clean_str(source_str, replace="?"):
    """Ensure :ref:`ASCII` characters, and replace any non ascii

    :param source_str: The source str to clean
    :type source_str: str
    :param replace: The character(s) to replace non ascii characters with
    :type replace: str
    :return: a `tuple` containing

             - The ascii string
             - `True` if a non ascii character was encounteres, otherwise `False`
    """
    illegal = False
    try:
        sss = source_str.decode("ascii")
        return sss, illegal
    except UnicodeDecodeError:
        safeS = ""
        illegal = True
        for cIdx in range(0, len(source_str)):
            singleS = source_str[cIdx]
            if ord(singleS) < 128:
                safeS += singleS
            else:
                safeS += replace
    return safeS, illegal


def delete_dir_contents(dir_path):
    """Deletes all the contents of a directory, not the directory itself

    :type dir_path: str
    :param dir_path: The relative or absolute path to the dir
    """
    filelist = glob.glob("%s/*" % dir_path)
    for f in filelist:
        pass  # os.remove(f)


def file_size_format(num, suffix='B'):
    """Formats a file size into human readable

    :param num: size (bytes)
    :param suffix:

        - **B** is default and return bytes
        - or one of **Ki**, **Mi**, **Gi**, **Ti**, **Pi**', **Ei**, **Zi**

    :return: **`str`** with file size
    """
    for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)


def file_size(file_path, human=True):
    """Returns a file's size

    :param file_path: relative or absolute path to file
    :type file_path: str
    :type human: bool
    :param human:

        - If **False**, return an **int** of the file size in bytes
        - If **True** returns a **str** with bytes in Kb, MB
    :return: `str` or `int` with the size
    """
    b = os.path.getsize(file_path)
    if human:
        return file_size_format(b)
    return b


def to_int(obj, default=None):
    try:
        return int(obj)
    except Exception as e:
        pass
    return default


def to_float(v, default=None):
    try:
        return float(v)
    except Exception as e:
        pass
    return default


def read_file(file_path):
    if not os.path.exists(file_path):
        return None, "file path '%s' not exist" % file_path

    with open(file_path, "r") as f:
        return f.read(), None
    return None, "Error reading '%' " % file_path


def write_file(file_path, contents):
    with open(file_path, "w") as f:
        f.write(contents)
        return None
    return "OOPS in write_file()"




import os
import sys
import yaml

HERE_PATH =  os.path.abspath( os.path.dirname( __file__ ))
PROJECT_ROOT = os.path.abspath( os.path.join(HERE_PATH, "..") )
if sys.path.count(PROJECT_ROOT) == 0:
    sys.path.insert(0, PROJECT_ROOT)


from werkzeug.utils import secure_filename
from flask import Flask, render_template, jsonify, request, flash, redirect


from ogt import ags4, ogt_doc


app = Flask(__name__,
            static_url_path="",
            static_folder=os.path.join(PROJECT_ROOT, "static"))



ALLOWED_UPLOADS = ['ags', 'ags4']


nav = [
    {"url": "/", "label": "Home"},
    #{"url": "/about", "label": "About"},
    {"url": "/ags4", "label": "AGS4 Data Dict"},
    {"url": "/viewer", "label":"Viewer"},
    {"url": "/convert", "label":"Convert"}
]

class Context(object):
    pass

def make_page_context(url, page_title):
    """Simple way to bang all vaiables into context"""
    c = Context()
    c.error = None
    c.site_name = "OGT Server"
    c.nav = nav
    c.url = url
    c.page_title = page_title
    return c
    """
    c = {"site_name": "OGT Server",
        "nav": nav,
        "url": url,
        "page_title": page_title
    }
    return c
    """

def xrender(ext, data):
    """Our custom function to knock out the stuff"""
    pass

@app.route('/')
def h_index():
    c = make_page_context("/", "OGT Server")
    return render_template('index.html', c=c)

@app.route('/about')
def h_about():
    c = make_page_context("/about", "About")
    return render_template('about.html', c=c)

@app.route('/viewer')
def h_viewer():
    c = make_page_context("/viewer", "Viewer")
    return render_template('viewer.html', c=c)


@app.route('/ags4')
def h_ags4_index():
    c = make_page_context("/ags4", "AGS4 Data Dict")
    return render_template('ags4.html', c=c)

@app.route('/ags4/<page>')
@app.route('/ags4/<page>/<xcode>')
def h_ags4_widget(page, xcode=None):

    tabs = [
        {"url": "groups", "label": "Groups"},
        {"url": "abbrs", "label": "Abbreviations"},
        {"url": "types_units", "label": "Types & Units"},
        #{"url": "data_types", "label": "Data Types"},
        {"url": "widget", "label": "Widget"},
    ]

    c = make_page_context("/ags4", "AGS4")
    c.tabs = tabs
    c.page = page

    tpl = "ags4.html"
    if page == "widget":
        tpl = 'ags4_widget.html'

    elif page == "abbrs":
        c.ags4_abbrs = ags4.DD.abbrs_dict
        tpl = "ags4_abbrs.html"

    elif page == "groups":
        c.ags4_classified_groups = ags4.DD.classified_groups()
        tpl = "ags4_groups.html"

    elif page == "group":
        c.page = "groups"
        c.ags4_group = ags4.DD.group(xcode)
        tpl = "ags4_group.html"

    elif page == "types_units":
        c.data_types = ags4.DD.data_types_list
        c.units = ags4.DD.units_list
        tpl = "ags4_types_units.html"


    return render_template(tpl, c=c)


@app.route('/ags4/groups.<ext>')
def ags4_groups(ext="json"):
    if ext == "json":
        return jsonify({"groups": ags4.groups(), "success": True})

    if ext in ["yml", "yaml"]:
        return yaml.dump(ags4.all())


@app.route('/ags4/groups_list.<ext>')
def ags4_groups_list(ext="json"):
    if ext == "json":
        return jsonify({"groups_list": ags4.DD.groups_dict().values(), "success": True})


@app.route('/ags4/group/<group_code>.<ext>')
def ags4_group(group_code, ext="json"):
    if ext == "json":
        return jsonify({"group": ags4.DD.group(group_code), "success": True})

    if ext in ["yml", "yaml"]:
        return yaml.dump(ags4.all())





@app.route('/ags4/abbrs_list.<ext>')
def ags4_abbrs_list(ext="html"):
    if ext == "json":
        return jsonify({"abbrs_list": ags4.abbrs(), "success": True})

    if ext in ["yml", "yaml"]:
        return yaml.dump(ags4.all())

    c = make_page_context("/ags4/groups", "AGS4 Groups")

    # TODO Make its nested based on class
    c['ags4_classified_groups'] = ags4.classified_groups()
    #print c['ags4_classified_groups']
    return render_template("ags4_groups.html", c=c)






@app.route('/ags4/examples/all')
@app.route('/ags4/examples/all.<ext>')
def ags4_examples_all(ext="html"):

    if ext == "json":
        return jsonify({"examples": ags4.examples(), "success": True})

    c = make_page_context("/ags4/examples", "AGS4 Widget")
    return render_template('ags4_examples.html', c=c)

@app.route('/ags4/examples')
def ags4_examples_page():
    """Examples list page"""
    c = make_page_context("/ags4/examples", "AGS4 Examples")
    c.examples, c.error = ags4.examples_list()
    return render_template('ags4_examples.html', c=c)

@app.route('/ags4/examples.json')
def ags4_examples_json():
    """Examples list json"""
    return jsonify({"examples": ags4.examples_list(), "success": True})


@app.route('/ags4/example')
@app.route('/ags4/example.<ext>')
def ags4_example(ext="html"):

    file_name = request.args.get('file_name')
    if not file_name:
        panic
    ex, err = ags4.example(file_name)
    if ex:
        #print "EX=", ex
        doc = ogt_doc.OGTDocument()

        #doc.source_file_path = ex['file_name']
        err = doc.add_ags4_string(ex['contents'], file_name)
        doc.opts.extended = True
        doc.opts.include_stats = False
        doc.opts.include_source = True
        #if request.args.get('format') == "json":
        #        return jsonify(doc.to_dict())
        if ext == "json":
            return jsonify({"document": doc.to_dict(), "success": True})

    c = make_page_context("/ags4/examples", "AGS4 Widget")
    return render_template('ags4_example.html', c=c)

def allowed_file(filename):
    return '.' in filename and filename.rsplit('.', 1)[1].lower() in ALLOWED_UPLOADS

@app.route('/convert', methods=["GET", "POST"])
def c_convert():

    c = make_page_context("/convert", "AGS4 Convert")
    c['err_mess'] = None
    if request.method == "POST":
        if 'ags_file' not in request.files:
            c['err_mess'] = "Need a file"

        else:
            file = request.files['ags_file']
            # if user does not select file, browser also
            # submit a empty part without filename
            if file.filename == '':
                c['err_mess'] = "Need a file"

            else:
                if file and allowed_file(file.filename):


                    doc = ogt_doc.OGTDocument()
                    doc.source_file_path = secure_filename(file.filename)
                    err = doc.load_ags4_string(file.read())

                    return jsonify(doc.to_dict())

    ## convert example
    """
    if request.method == "GET":
        ex_name = request.args.get('example')
        if ex_name:

            ex = ags4.example(ex_name)
            #print "EX=", ex
            doc = ogt_doc.OGTDocument()
            doc.source_file_path = ex_name
            err = doc.load_ags4_string(ex['contents'])

            if request.args.get('format') == "json":
                return jsonify(doc.to_dict())
    """

    return render_template('convert.html', c=c)

"""
@app.route('/convert.json', methods=["GET", "POST"])
def ajax_convert():

    ex_name = request.args.get('example')
    if ex_name:
        ex = ags4.example(ex_name)
        #print "EX=", ex
        doc = ogt_doc.OGTDocument()
        doc.source_file_path = ex_name
        err = doc.load_ags4_string(ex['contents'])

        return jsonify(doc.to_dict())
"""



# -*- coding: utf-8 -*-
"""
@author: Peter Morgan <pete@daffodil.uk.com>
"""
import json
from Qt import QtNetwork, QtCore, Qt, pyqtSignal

import app_globals as G

ATTR_ID = QtNetwork.QNetworkRequest.User


class XReply:
    """This is the data that is emitted request finished et all"""
    def __init__(self, qreply, tid=None, origin=None, status=None, flag=None, error=None, data=None):

        self.origin = qreply.request().originatingObject()
        self.http_code = qreply.attribute(QtNetwork.QNetworkRequest.HttpStatusCodeAttribute)
        self.tid = str(qreply.request().attribute(ATTR_ID))

        self.url = qreply.request().url().toString(QtCore.QUrl.RemoveScheme|QtCore.QUrl.RemovePort)


        self.error = error

        self.data = None



    def __repr__(self):
        d = "-"
        if self.data:
            if isinstance(self.data, dict):
                d = sorted(self.data.keys())
            else:
                d = self.data
        return "<XReply ori=%s, err=%s data=%s>" % (self.origin.__class__.__name__,  self.error, d)

    def debug(self):
        if self.data:
            print(self.data.keys())


class ServerConnection( QtCore.QObject ):
    """
    HTTP Client
    """
    TIMEOUT = 3000 #30000

    response = pyqtSignal(object) # XReply

    def __init__( self, parent=None, server=None ):
        QtCore.QObject.__init__( self, parent )

        self.debug = False

        self.callbacks = {}

        self.netManager = QtNetwork.QNetworkAccessManager( self )

        ## Initialise cookies
        self.cookieJar = QtNetwork.QNetworkCookieJar()
        self.netManager.setCookieJar( self.cookieJar )


        self.netManager.finished.connect(self.on_request_finished)
        #self.netMan.error.connect(self._on_server_read_error)



    def get( self, origin=None, url=None, params=None, tag=None,
             cb=None, widget=None,  debug=False, spinner=True):


        self.debug = False


        srv = G.settings.current_server()
        if srv == None:
            print("NO Server in dServer.fetch()")
            return

        url = QtCore.QUrl( "%s/ajax%s" % ( srv.url, url ) )

        #q = QtCore.QUrlQuery()


        if params:
            for k, v in params.items():
                url.addQueryItem( str( k ), str( v ) )

        request = QtNetwork.QNetworkRequest()
        request.setUrl( url )
        request.setOriginatingObject(origin)
        request.setPriority(QtNetwork.QNetworkRequest.HighPriority)
        if tag:
            request.setAttribute(ATTR_ID, tag)

        #self.load_cookies() TODO

        if spinner:
            pass #self.spin(True)


        if G.args.dev:
            print("# --- GET: tag=%s, %s" % (tag, url.toString()))
        #self.trigger_reply( request, SERVER_STATUS.REQUESTING, SERVER_FLAG.WAIT )
        ## creat own accoutn bundle and senf

        #req = Request(url)
        ##req.timeout.connect(self.on_timeout)
        ##self.requests.append(req)
        #req.reply = self.netManager.get( request )
        reply = self.netManager.get( request )



    def on_request_finished( self, qreply ):

        """Server Request has finished, so parse and check for errors"""
        #self.spin(False)

        reply = XReply(qreply=qreply)
        reply.tag = qreply.request().attribute(ATTR_ID)

        reply.origin = qreply.request().originatingObject()

        ## Things not OK
        if reply.http_code != 200:
            reply.error = "HTTP Error: %s" % reply.http_code
            qreply.deleteLater()
            self.response.emit(reply)
            return

        if qreply.error():
            print("ERROR", qreply.error(), qreply.attribute(QtNetwork.QNetworkReply.HttpStatusCodeAttribute))
            catchme_todo
            qreply.deleteLater()
            return

        ## Save Cookies - TODO
        """
        cookies = self.netManager.cookieJar().allCookies()
        #print "-------- REC COOKIES----------------_"
        G.settings.beginGroup( "cookies" )
        for cookie in cookies:
            G.settings.setValue( "%s" % str( cookie.name() ), str( cookie.value() ) )
            #print "SAVE", cookie.name(), cookie.value()
        G.settings.endGroup()
        """


        ## Decode json
        # make it py/str/ascii for now (need to catch non ascii soon and utf-8 everywhere)
        # TODO catch non unicode
        contents = str(qreply.readAll().data())
        #print("contents=", contents)

        try:
            reply.data = json.loads(contents)

            # Always expect a dict or list so add error
            if isinstance(reply.data, dict) or isinstance(reply.data, list):
                pass # ok
            else:
                reply.error = "Does nto seem to be json"

            qreply.deleteLater()
            self.response.emit(reply)
            return

        except ValueError as e:
            reply.error = "Not Valid json %s" % str(e)
            qreply.deleteLater()
            self.response.emit(reply)
            return

        except TypeError as e:
            reply.error = "Not Valid json %s" % str(e)
            qreply.deleteLater()
            self.response.emit(reply)
            return

        ## json is decoded and works
        print("should not get here")
        if self.debug:
            print("       data keys: %s" % " ".join( sorted( resp.data.keys() ) ))
        qreply.deleteLater()



# -*- coding: utf-8 -*-


from Qt import QtGui, QtCore, Qt, pyqtSignal

from ogt import ags4

from ogtgui.img  import Ico
from ogtgui import xobjects




class AGS4_TYPE:
    abbrev = "ABBR"
    abbrev_item = "ABBR_ITEM"
    group = "GROUP"
    heading = "HEADING"
    note = "NOTE"


class AGS4_COLORS:
    group = "#286225"
    abbrev = "#496FA3"


SHOW_NONE = "##__NONE__##"
"""Used to filter for nothing"""

def data_type_ico(ags_type):
    """Returns an icon filename for the data type"""
    v = ags_type.upper()

    # Date time
    if v in ["DT"]:
        return Ico.TypeDate

    # decimal places
    if v.endswith("DP"):
        return  Ico.TypeDecimal

    # UID
    # TODO: (is this realy UID in doucment ?? aslks pete)
    if v == "ID":
        return  Ico.TypeID

    # Picklists
    if v in ["PA", "PU", "PT"]:
        return  Ico.TypePicklist

    # Scientific stuff
    if v.endswith("SCI"):
        return  Ico.TypeSci

    # Boolean yes/no, false, true, not false
    if v == "YN":
        return  Ico.TypeCheckBox

    # text and text in general
    if v in ["X"]:
        return  Ico.TypeText

    # oopps
    return Ico.TypeUnknown

def data_type_qicon(ags_type):
    """Return a QIcon for and ags.data_type"""
    return Ico.icon(data_type_ico(ags_type))

##===================================================================
## Main
##===================================================================
class Ags4Object(QtCore.QObject):

    sigLoaded = pyqtSignal()

    def __init__( self, parent=None):
        super(QtCore.QObject, self).__init__(parent)

        self.modelUnits = UnitsModel(self)
        self.modelDataTypes = DataTypesModel(self)

        self.modelClasses = ClassesModel(self)

        self.modelGroups = GroupsModel(self)

        self.modelHeadings = HeadingsModel(self)

    def init(self):
        #self.modelNotes = NotesModel()
        self.units = UnitsModel(self)
        self.dataTypes = DataTypesModel(self)

        self.classes = ClassesModel(self)

        self.headings = HeadingsModel(self)

        #self.modelAbbrevClasses = ClassesModel(self)
        #self.modelAbbrevs = AbbrevsModel(self)

        #self.modelAbbrItems = AbbrevItemsModel(self)

        #self.modelGroups.sigClasses.connect(self.modelClasses.load_classes)

        #self.connect(self.modelAbbrevs, QtCore.SIGNAL("classes"), self.modelAbbrevClasses.load_classes)


    def load_init(self):
        #err = ags4.DD.setup_initialise()
        #if err:
        #    panicssss

        self.modelClasses.set_ags4dd(ags4.DD)
        self.modelGroups.set_ags4dd(ags4.DD)

        self.modelUnits.set_ags4dd(ags4.DD)
        #self.modelDataTypes.set_ags4dd(ags4.DD)

        #self.modelNotes.init_words()

        self.sigLoaded.emit()



    def get_group(self, code):
        return self.modelGroups.get_group(code)


    def get_notes(self, group_code):
        return self.modelNotes.get_notes(group_code)

    def get_heading(self, head_code):
        return self.modelHeadings.get_heading(head_code)

    def DEADget_picklist(self, abbrev):

        return self.modelAbbrevItems.get_picklist(abbrev)

class CG:
    code = 0
    description = 1
    cls = 2
    search = 3
    _col_count = 4

##===================================================================
## Groups
##===================================================================
class GroupsModel(QtCore.QAbstractItemModel):

    def __init__( self, parent=None):
        QtCore.QAbstractItemModel.__init__( self, parent)
        self.ags4dd = None

    def set_ags4dd(self, ags4dd):
        self.ags4dd = ags4dd
        #self.reset()

    def columnCount(self, midx):
        return CG._col_count

    def rowCount(self, midx):
        if self.ags4dd == None:
            return 0
        return self.ags4dd.groups_count()

    def index(self, row, col, parent=None):
        return self.createIndex(row, col)

    def parent(self, pidx=None):
        return QtCore.QModelIndex()

    def data(self, midxx, role):

        col = midxx.column()

        if role == Qt.DisplayRole:
            # the the ags grp
            grp = self.ags4dd.group_by_index(midxx.row())

            if col == CG.code:
                return grp['group_code']
            if col == CG.description:
                return grp['group_description']
            if col == CG.cls:
                return grp['class']
            if col == CG.search:
                return (grp['group_code'] + grp['group_description']).replace(" ", "")

        if role == Qt.DecorationRole:
            if col == CG.code:
                return Ico.icon(Ico.AgsGroup)

        if role == Qt.ForegroundRole:
            if col == CG.code:
                grp = self.ags4dd.group_by_index(midxx.row())
                return QtGui.QColor("#990000" if grp['group_required'] else "#000000")

        if role == Qt.FontRole:
            if col == CG.code:
                f = QtGui.QFont()
                f.setBold(True)
                f.setFamily("monospace")
                return f

        return None

    def rec_from_midx(self, midx):
        return self.ags4dd.group_by_index(midx.row())

    def headerData(self, p_int, orient, role=None):

        if orient == Qt.Horizontal and role == Qt.DisplayRole:
            heads = ["Code", "Description", "Classification", "Search"]
            return heads[p_int]


    def get_group(self, code):
        items = self.findItems(code, Qt.MatchExactly, CG.code)
        if len(items) == 0:
            # not in ags data dict
            return None
        ridx = items[0].index().row()
        return dict(	group_code=self.item(ridx, CG.code).s(),
                        group_description=self.item(ridx, CG.description).s(),
                        cls=self.item(ridx, CG.cls).s())

    def get_words(self):

        lst = []
        for ridx in range(0, self.rowCount()):
            lst.append( dict(type=AGS4_TYPE.group, description=self.item(ridx, CG.description).s(),
                            code=self.item(ridx, CG.code).s()))
        return lst


##===================================================================
## Headings
##===================================================================

class HeadingsModel(QtCore.QAbstractTableModel):

    class CH:
        """Columns NO's for the ;class:`~ogtgui.ags_models.HeadingsModel`"""
        action = 0
        head_code = 1
        description = 2
        unit = 3
        data_type = 4
        key = 5
        required = 6
        sort_order = 7
        example = 8
        _col_count = 9

    def __init__( self, parent):
        QtCore.QAbstractTableModel.__init__(self, parent)

        self.grpDD = None


    def index(self, row, col, parent=None):
        return self.createIndex(row, col)

    def set_group(self, grpdd):

        self.grpDD = grpdd

        self.modelReset.emit()

    def columnCount(self, midx=None):
        return self.CH._col_count

    def rowCount(self, midx=None):
        if self.grpDD == None:
            return 0
        return len(self.grpDD.get("headings_index"))

    def index(self, row, col, parent=None):
        return self.createIndex(row, col)

    def parent(self, pidx=None):
        return QtCore.QModelIndex()

    def data(self, midxx, role):

        col = midxx.column()

        if role == Qt.FontRole:
            if col == self.CH.head_code:
                f = QtGui.QFont()
                f.setBold(True)
                f.setFamily("monospace")
                return f


        if role == Qt.DisplayRole:

            rec = self.rec_from_midx(midxx)

            if col == self.CH.head_code:
                return rec['head_code']

            if col == self.CH.description:
                return rec['head_description']

            if col == self.CH.data_type:
                return rec['data_type']

            if col == self.CH.unit:
                return rec['unit']

            #if col == self.CH.key:
            #    return "KEY" if rec['key'] else ""

            if col == self.CH.required:
                return "Req" if rec['required'] else ""

            if col == self.CH.example:
                return rec['example']

            if col == self.CH.sort_order:
                return midxx.row() #rec['sort_order']

            return None

        if role == Qt.DecorationRole:
            if col == self.CH.data_type:
                rec = self.rec_from_midx(midxx)
                return data_type_qicon( rec['data_type'] )#

            if col == self.CH.key:
                rec = self.rec_from_midx(midxx)
                if rec['key']:
                    return Ico.icon(Ico.KeyField)

        if role == Qt.BackgroundRole:
            if col == self.CH.head_code:
                rec = self.rec_from_midx(midxx)
                if rec['head_code'].split("_")[0] == self.grpDD['group_code']:
                    return QtGui.QColor("#D2FBC4")
                #return QtGui.QColor("#efefef")

            if col == self.CH.required:
                rec = self.rec_from_midx(midxx)
                if rec['required']:
                    return QtGui.QColor("pink")
                #return QtGui.QColor("transparent")

        return None

    def get_rec(self, idx):
       return self.grpDD["headings"][ self.grpDD.get("headings_index")[idx] ]

    def get_heading_index(self, code):
        for ridx, rec in enumerate(self.grpDD["headings"]):
            if rec['head_code'] == code:
                return self.index(ridx, 0)
        return None

    def rec_from_midx(self, midx):
        return self.grpDD["headings"][ midx.row() ]

    def headerData(self, p_int, orient, role=None):

        if orient == Qt.Horizontal and role == Qt.DisplayRole:
            heads = ["", "", "Heading", "Description", "Unit", "Type", "Key","Req", "Srt", "Example"]
            return heads[p_int]


##===================================================================
## Abbrev Values
##===================================================================

class AbbrevItemsModel(QtCore.QAbstractItemModel):

    class MODE:
        doc = 0
        ags = 1


    class CA:
        """Columns no's for the :class:`~ogtgui.ags4_models.AbbrevItemsModel` """
        check = 0
        code = 1
        description = 2
        list = 3
        _col_count = 4

    CHK_KI = "__CHECKED__"

    sigCheckedChanged = pyqtSignal()

    def __init__( self, parent=None, ogtHead=None):
        QtCore.QAbstractItemModel.__init__(self, parent)

        self.ogtHead = ogtHead
        self.recs = None

        self.load_data()

    def load_data(self):
        if self.ogtHead == None:
            self.layoutChanged.emit()
            return

        self.recs = self.ogtHead.abbrs()
        #print(self.recs)
        self.modelReset.emit()

    def set_ags_heading(self, head):

        self.recs = None
        if head:
            abbrs = ags4.DD.abbrs_dict.get(head['head_code'])
            if abbrs:
                self.recs = abbrs
        self.modelReset.emit()

    def set_checked(self, selected):

        sel_code = selected['code'] if isinstance(selected, dict) else selected

        for i in range(0, self.rowCount()):
            rec = self.recs[i]
            rec[self.CHK_KI] = rec['code'] == sel_code

        self.layoutChanged.emit()

    def get_checked(self):
        for rec in self.recs:
            if rec.get(self.CHK_KI) == Qt.Checked:
                return rec
        return None

    def flags(self, midx):
        if midx.column() == self.CA.check:
            return Qt.ItemIsUserCheckable | Qt.ItemIsEnabled | Qt.ItemIsEditable | Qt.ItemIsSelectable
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable

    def index(self, row, col, parent=None):
        return self.createIndex(row, col)

    def parent(self, pidx=None):
        return QtCore.QModelIndex()

    def columnCount(self, midx=None):
        return self.CA._col_count

    def rowCount(self, midx=None):
        if self.recs == None:
            return 0
        return len(self.recs)

    def headerData(self, p_int, orient, role=None):
        if orient == Qt.Horizontal and role == Qt.DisplayRole:
            heads = ["", "Code", "Description", "List"]
            return heads[p_int]

    def data(self, midxx, role):

        col = midxx.column()

        if role == Qt.DisplayRole:
            rec = self.recs[midxx.row()]

            #if col == self.CA.check:
            #    return rec['code']

            if col == self.CA.code:
                return rec['code']

            if col == self.CA.description:
                return rec['description']

            if col == self.CA.list:
                return rec['list']


        # checked
        if col == self.CA.check and role == Qt.CheckStateRole:
            return Qt.Checked if self.recs[midxx.row()].get(self.CHK_KI) else Qt.Unchecked

        # Background color
        if role == Qt.BackgroundRole:
            rec = self.recs[midxx.row()]
            if col == self.CA.list and rec["list"] != "AGS4":
                return QtGui.QColor("#FFCF9A")

            if rec.get(self.CHK_KI):
                if col == self.CA.check:
                    return QtGui.QColor("darkgreen")
                return QtGui.QColor("#FCFF94")
            return QtGui.QColor("transparent")

        # Make code font monospaced out man
        if col == self.CA.code and role == Qt.FontRole:
            f = QtGui.QFont()
            f.setBold(True)
            f.setFamily("monospace")
            return f

        return None

    def checked_count(self):
        return len(self.checked_codes())

    def checked_codes(self):
        lst = []
        for rec in self.recs:
            if rec[self.CHK_KI] and not rec['code'] in lst:
                lst.append(rec['code'])
        return sorted(lst)

    def midx_from_val(self, v):

        for ridx, rec in enumerate(self.recs):
            if rec['code'] == v:
                return self.index(ridx, 0)
        return None

    def val_from_midx(self, midx):
        return self.recs[midx.row()]


    def setData(self, midx, vv, role):

        if role == Qt.CheckStateRole and midx.column() == self.CA.check:

            # clear existing checked
            for rec in self.recs:
                rec[self.CHK_KI] = Qt.Unchecked

            rec = self.recs[midx.row()]
            rec[self.CHK_KI] = vv
            self.sigCheckedChanged.emit()
        self.layoutChanged.emit()
        return True

    def set_save_cell(self):
        #lst = []
        #for rec in self.recs:
        #    if rec.get(self.CHK_KI) and not rec['code'] in lst:
        #        lst.append(rec['code'])
        #
        cc = self.ogtHead.parentGroup.parentDoc.concat_char
        return cc.join( sorted(self.checked_codes()) )


##===================================================================
## Classes
##===================================================================
class ClassesModel(xobjects.XStandardItemModel):

    def __init__( self, parent=None):
        super(xobjects.XStandardItemModel, self).__init__(parent)

        self.ags4dd = None



    def set_ags4dd(self, ags4dd):
        self.ags4dd = ags4dd

        self.set_header(0, "Class")

        ## make root node,
        # all "class label" are kidz.
        items = self.make_blank_row()
        items[0].set("All", ico=Ico.Folder)
        self.appendRow(items)

        ## get unique list of classes by walking groups
        classes = []
        for g in self.ags4dd.groups_list():
            if not g['class'] in classes:
                classes.append(g['class'])
        classes.sort()

        rootItem = self.item(0, 0)  # 'All' parent

        ## remove existing nodes
        while rootItem.hasChildren():
            rootItem.removeRow(0)

        ## add items
        for r in classes:
            citems = self.make_blank_row()
            citems[0].set(r)
            rootItem.appendRow(citems)


##===================================================================
## Units + Types
##===================================================================

class UnitsModel(QtCore.QAbstractItemModel):

    class CU:
        unit = 0
        description = 1

    def __init__( self, parent=None):
        super(QtCore.QAbstractItemModel, self).__init__(parent)

        self.ags4dd = None

    def set_ags4dd(self, ags4dd):
        self.ags4dd = ags4dd
        self.layoutChanged.emit()

    def index(self, row, col, parent=None):
        return self.createIndex(row, col)

    def parent(self, pidx=None):
        return QtCore.QModelIndex()

    def columnCount(self, midx=None):
        return 2

    def rowCount(self, midx=None):
        if self.ags4dd == None:
            return 0
        return len(self.ags4dd.units_list)

    def headerData(self, p_int, orient, role=None):
        if orient == Qt.Horizontal and role == Qt.DisplayRole:
            heads = ["Type", "Description"]
            return heads[p_int]

    def data(self, midxx, role):

        col = midxx.column()

        if role == Qt.DisplayRole:
            # the the ags grp
            rec = self.ags4dd.units_list[midxx.row()]
            if col == self.CU.unit:
                return rec['code']
            if col == self.CU.description:
                return rec['description']

        if role == Qt.FontRole:
            if col == self.CU.unit:
                f = QtGui.QFont()
                f.setBold(True)
                f.setFamily("monospace")
                return f

        return None

    def midx_from_val(self, v):

        for ridx, rec in enumerate(self.ags4dd.units_list):
            if rec['code'] == v:
                return self.index(ridx, 0)
        return None

    def val_from_midx(self, midx):
        return self.ags4dd.units_list[midx.row()]


class DataTypesModel(QtCore.QAbstractItemModel):

    class CT:
        data_type = 0
        description = 1
        _col_count = 2



    def __init__( self, parent=None):
        super().__init__(parent)




    def midx_from_val(self, v):

        for ridx, rec in enumerate(ags4.DD.data_types):
            if rec['code'] == v:
                return self.index(ridx, 0)
        return None

    def val_from_midx(self, midx):
        return ags4.DD.data_types[midx.row()]

    def index(self, row, col, parent=None):
        return self.createIndex(row, col)

    def parent(self, pidx=None):
        return QtCore.QModelIndex()

    def columnCount(self, midx=None):
        return self.CT._col_count

    def rowCount(self, midx=None):
        if ags4.DD.data_types_list == None:
            return 0
        return len(ags4.DD.data_types_list)

    def headerData(self, p_int, orient, role=None):
        if orient == Qt.Horizontal and role == Qt.DisplayRole:
            heads = ["Type", "Description"]
            return heads[p_int]

    def data(self, midxx, role):

        col = midxx.column()

        if role == Qt.DisplayRole:
            rec = ags4.DD.data_types[midxx.row()]
            if col == self.CT.data_type:
                return rec['code']
            if col == self.CT.description:
                return rec['description']

        # if role == Qt.DecorationRole:
        #     if col == 0:
        #         return Ico.icon(Ico.AgsHeading)

        if role == Qt.FontRole:
            if col == 0:
                f = QtGui.QFont()
                f.setBold(True)
                f.setFamily("monospace")
                return f

        return None



AgsData = Ags4Object()



# -*- coding: utf-8 -*-


import os
import collections

from Qt import QtGui, QtWidgets, QtCore, Qt, pyqtSignal

import xwidgets
from img import Ico
from ogt import ags4
from ogt.ogt_validate import OGTError
#from ogt import CELL_COLORS, HAVE_EXCEL
import ogtgui_excel
import ogtgui_errors



import app_globals as G

def bg_color(descr):
    if descr == ags4.AGS4.GROUP:
        return "#D4C557"

    if descr == ags4.AGS4.HEADING:
        return "#FCF66D"

    if descr in [ags4.AGS4.UNIT, ags4.AGS4.TYPE]:
        return "#FFE8B9"

    if descr == ags4.AGS4.DATA:
        return "#DFD1FF"

    return "#ffffff"

class FILTER_ROLE:
    warn = Qt.UserRole + 3
    err = Qt.UserRole + 5


class C_EG:
    """Columns for examples"""
    file_name = 0

class ExamplesWidget( QtWidgets.QWidget ):

    sigFileSelected = pyqtSignal(object)

    def __init__( self, parent):
        QtWidgets.QWidget.__init__( self, parent )

        self.debug = False

        self.setMinimumWidth(300)

        self.mainLayout = QtWidgets.QVBoxLayout()
        self.mainLayout.setSpacing(0)
        self.mainLayout.setContentsMargins(0,0,0,0)
        self.setLayout(self.mainLayout)


        #=============================
        ## Set up tree
        self.tree = QtWidgets.QTreeWidget()
        self.mainLayout.addWidget(self.tree, 30)

        self.tree.setRootIsDecorated(False)
        self.tree.header().setStretchLastSection(True)
        self.tree.header().hide()

        hi = self.tree.headerItem()
        hi.setText(C_EG.file_name, "Example")


        self.tree.itemClicked.connect(self.on_tree_item_clicked)

        self.load_files_list()



    def load_files_list(self, sub_dir=None):

        files_list, err = ags4.examples_list()
        if err:
            pass #TODO
        self.tree.clear()
        if files_list == None:
            return
        for fd in files_list:
            file_name = fd["file_name"]
            item = QtWidgets.QTreeWidgetItem()
            item.setText(C_EG.file_name, file_name)
            item.setIcon(C_EG.file_name, Ico.icon(Ico.Ags4))
            f = item.font(C_EG.file_name)
            f.setBold(True)
            item.setFont(C_EG.file_name, f)
            self.tree.addTopLevelItem(item)



    def on_tree_item_clicked(self, item, col):

        file_name = str(item.text(C_EG.file_name))
        self.sigFileSelected.emit(file_name)


class ErrorsListSortFilterModel(QtCore.QSortFilterProxyModel):

    def __init__(self):
        QtCore.QSortFilterProxyModel.__init__(self)

        self.show_errors = True
        self.show_warnings = True


    def dfilterAcceptsRow(self, ridx, midx):
        pass





class HelpDialog(QtWidgets.QDialog):
    debug = False

    def __init__(self, parent=None, page=None):
        QtWidgets.QDialog.__init__(self, parent)

        self.setWindowTitle("Help")
        self.setWindowIcon(dIco.icon(dIco.Help))

        # self.setWindowFlags(QtCore.Qt.Popup)
        self.setWindowModality(QtCore.Qt.ApplicationModal)

        self.setMinimumWidth(900)
        self.setMinimumHeight(700)

        # self.setStyleSheet("border: 1px solid black;")

        outerLayout = QtWidgets.QVBoxLayout()
        outerLayout.setSpacing(0)
        margarine = 0
        outerLayout.setContentsMargins(margarine, margarine, margarine, margarine)
        self.setLayout(outerLayout)

        mainLayout = QtWidgets.QHBoxLayout()
        mainLayout.setSpacing(0)
        margarine = 0
        mainLayout.setContentsMargins(margarine, margarine, margarine, margarine)
        outerLayout.addLayout(mainLayout)

        self.helpWidget = HelpWidget.HelpWidget(page=page)
        mainLayout.addWidget(self.helpWidget)

    # if page:
    #	self.show_page(page)

    def show_page(self, page):
        self.helpWidget.load_page(page)
        self.exec_()



#import mistune
#from help import DOCS_PATH, DOCS_CONTENT

class C:
    node = 0
    page = 1


class HelpWidget(QtWidgets.QWidget):
    def __init__(self, parent=None, page=None):
        QtWidgets.QWidget.__init__(self, parent)

        self.debug = False

        outerLayout = QtWidgets.QVBoxLayout()
        outerLayout.setSpacing(0)
        margarine = 0
        outerLayout.setContentsMargins(margarine, margarine, margarine, margarine)
        self.setLayout(outerLayout)

        self.lblTitle = QtWidgets.QLabel()
        self.lblTitle.setText("Help")
        self.lblTitle.setStyleSheet("background-color: #333333; color: #999999; font-size: 19pt; padding: 4px; ")
        outerLayout.addWidget(self.lblTitle, 0)

        outerLayout.addSpacing(5)

        self.splitter = QtWidgets.QSplitter()
        outerLayout.addWidget(self.splitter, 100)

        ## Contents Tree
        self.tree = QtWidgets.QTreeWidget()
        self.splitter.addWidget(self.tree)
        hi = self.tree.headerItem()
        hi.setText(C.node, "File")
        hi.setText(C.page, "Page")

        self.tree.setColumnHidden(C.page, not self.debug)
        self.tree.header().setStretchLastSection(True)
        self.tree.setUniformRowHeights(True)

        self.tree.setFixedWidth(200)
        self.tree.itemSelectionChanged.connect(self.on_tree_selection)

        ## Tab Widget
        self.tabWidget = QtWidgets.QTabWidget()
        self.splitter.addWidget(self.tabWidget)
        self.tabWidget.setTabsClosable(True)
        self.tabWidget.tabCloseRequested.connect(self.on_tab_close_requested)
        self.tabWidget.currentChanged.connect(self.on_tab_changed)

        self.splitter.setStretchFactor(0, 1)
        self.splitter.setStretchFactor(1, 2)

        self.load_content_tree()

    def load_content_tree(self):

        self.tree.clear()
        self.tree.setUpdatesEnabled(False)

        item = TreeWidgetItem()
        item.set(C.node, "Index", ico=dIco.HelpPage)

        item.setText(C.page, "index.md")
        self.tree.addTopLevelItem(item)

        rootNode = self.tree.invisibleRootItem()
        self.load_dir_node(rootNode, DOCS_CONTENT)

        self.tree.setUpdatesEnabled(True)

    def load_dir_node(self, pItem, pth):

        files = os.listdir(pth)

        for f in sorted(files):
            xpth = os.path.join(pth, f)[len(DOCS_CONTENT):]
            if f == "index.md" and pth == DOCS_CONTENT:
                continue
            if os.path.isdir(f):
                item = TreeWidgetItem(pItem)
                item.set(C.node, f, ico=dIco.Folder)
                item.setText(C.page, xpth)

                self.load_dir_node(item, os.path.join(pth, f))
                self.tree.setItemExpanded(item, True)

            elif f.endswith(".md"):
                item = TreeWidgetItem(pItem)
                item.set(C.node, self.title_from_filename(f), ico=dIco.HelpPage)
                item.setText(C.page, xpth)

    def dpage_path_fn(self, page):
        if page.endswith(".md"):
            "%s.md" % page
        read_fn = os.path.join(DOCS_CONTENT, page)
        # if self.debug:
        if not os.path.exists(read_fn):
            print("ERROR: FILE not exits", read_fn, self)
            return

    def title_from_filename(self, fn):
        fn = os.path.basename(fn)
        if fn.endswith(".md"):
            fn = fn[0:-3]
        return fn.replace("_", " ").title()

    def select_page(self, page):

        idx = None
        if self.tabWidget.count() == 0:
            return idx
        ## check page is aleady loaded, and select
        for i in range(0, self.tabWidget.count()):

            if self.tabWidget.widget(i).page == page:
                self.tabWidget.blockSignals(True)
                self.tabWidget.setCurrentIndex(i)
                self.tabWidget.blockSignals(False)
                idx = i

                break
        items = self.tree.findItems(page, Qt.MatchExactly | Qt.MatchRecursive, C.page)
        if len(items) > 0:
            self.tree.blockSignals(True)
            self.tree.setCurrentItem(items[0])
            self.tree.blockSignals(False)
        return idx

    def load_page(self, page):

        page = str(page)

        full_path = DOCS_CONTENT + page

        if os.path.isdir(full_path):
            return
        elif not page.endswith(".md"):
            full_path = full_path + ".md"

        idx = self.select_page(page)
        if idx != None:
            return

        if not os.path.exists(full_path):
            # print "ERROR: FILE not exits", full_path
            return

        container = G.ut.read_file(os.path.join(DOCS_PATH, "templates", "help_container.html"))

        txt = G.ut.read_file(full_path)
        html = mistune.markdown(txt, escape=False)

        out_html = container.replace("##++CONTENT++##", html)

        self.tabWidget.blockSignals(True)
        webView = HelpPageView()
        nidx = self.tabWidget.addTab(webView, self.title_from_filename(page))
        webView.set_data(page, out_html)
        self.tabWidget.setTabIcon(nidx, dIco.icon(dIco.HelpPage))

        webView.sigPageLinkClicked.connect(self.load_page)

        self.tabWidget.setCurrentIndex(nidx)
        self.select_tree_node(page)
        self.tabWidget.blockSignals(False)

    def on_tree_selection(self):
        item = self.tree.currentItem()
        if item == None:
            return
        self.load_page(item.s(C.page))

    def on_tab_close_requested(self, idx):
        self.tabWidget.removeTab(idx)

    def on_tab_changed(self, nidx):
        if nidx == -1:
            self.tree.blockSignals(True)
            self.tree.clearSelection()
            self.tree.blockSignals(False)
            return
        page = self.tabWidget.widget(nidx).page
        self.select_tree_node(page)

    def select_tree_node(self, page, block=True):

        self.tree.blockSignals(True)
        items = self.tree.findItems(page, Qt.MatchExactly | Qt.MatchRecursive, C.page)
        if len(items) > 0:
            self.tree.setCurrentItem(items[0])
        else:
            # page not in menu
            pass
        self.tree.blockSignals(False)


class HelpPageView(QtWidgets.QWidget):
    sigPageLinkClicked = pyqtSignal(str)

    def __init__(self, parent=None, page=None):
        QtWidgets.QWidget.__init__(self, parent)

        self.debug = False
        self.page = None

        lay = QtWidgets.QVBoxLayout()
        lay.setSpacing(0)
        lay.setContentsMargins(0, 0, 0, 0)
        self.setLayout(lay)

        self.webView = QtWebKit.QWebView()
        lay.addWidget(self.webView, 2)

        if self.debug:
            page = self.webView.page()
            page.settings().setAttribute(QtWebKit.QWebSettings.DeveloperExtrasEnabled, True)
            # elf.webView.settings().globalSettings().setAttribute(QtWebKit.QWebSettings.DeveloperExtrasEnabled, True)

            self.webInspector = QtWebKit.QWebInspector(self)
            self.webInspector.setPage(page)
            lay.addWidget(self.webInspector, 3)

        self.webView.page().setLinkDelegationPolicy(QtWebKit.QWebPage.DelegateAllLinks)
        self.webView.linkClicked.connect(self.on_link_clicked)

    def set_data(self, page, html):
        self.page = page
        base_url = QtCore.QUrl.fromLocalFile(DOCS_PATH + "/")
        # bu.setScheme("file")

        self.webView.setHtml(html, base_url)

    def on_link_clicked(self, url):
        page = str(url.path())[1:]

        self.sigPageLinkClicked.emit(page)




class ExpPortalWidget( QtWidgets.QWidget ):
    """The SourceViewWidget info which in row 0 """



    def __init__( self, parent=None):
        QtWidgets.QWidget.__init__( self, parent )

        self.debug = False
        self.setObjectName("OGTSourceViewWidget")

        self.mainLayout = QtWidgets.QVBoxLayout()
        self.mainLayout.setSpacing(0)
        self.mainLayout.setContentsMargins(0,0,0,0)
        self.setLayout(self.mainLayout)


        self.tabWidget = QtWidgets.QTabWidget()
        self.mainLayout.addWidget(self.tabWidget)

        self.excelBrowse = ogtgui_excel.ExpExcelBrowserWidget()
        self.tabWidget.addTab(self.excelBrowse, "Excel Browse")

        self.pasteText = ExpPasteWidget()
        self.tabWidget.addTab(self.pasteText, "Paste clipboard from Excel")


    def init_load(self):
        pass

TEST_TXT = """
SAMP_TOP	SAMP_REF	SAMP_TYPE
m		
2DP	X	PA
4.50		U
"""



class ExpPasteWidget( QtWidgets.QWidget ):
    """The SourceViewWidget info which in row 0 """



    def __init__( self, parent=None):
        QtWidgets.QWidget.__init__( self, parent )

        self.debug = False
        self.setObjectName("OGTSourceViewWidget")

        self.mainLayout = QtWidgets.QVBoxLayout()
        self.mainLayout.setSpacing(0)
        self.mainLayout.setContentsMargins(0,0,0,0)
        self.setLayout(self.mainLayout)


        #self.tabWidget = QtWidgets.QTabWidget()
        #self.mainLayout.addWidget(self.tabWidget)

        self.txtSrc = QtWidgets.QPlainTextEdit()
        self.mainLayout.addWidget(self.txtSrc)
        self.txtSrc.textChanged.connect(self.on_text_changed)

        self.table = QtWidgets.QTableWidget()
        self.mainLayout.addWidget(self.table)
        self.txtSrc.setPlainText(TEST_TXT)


    def init_load(self):
        pass

    def on_text_changed(self):
        s = str(self.txtSrc.toPlainText())
        lines = s.split("\n")

        ridx = 0
        for rridx, rline in enumerate(lines):
            line = rline.strip()
            if line == "":
                continue
            if self.table.rowCount() < ridx + 1:
                self.table.setRowCount(ridx + 1)
            cols = line.split("\t")
            if self.table.columnCount() < len(cols):
                self.table.setColumnCount(len(cols))
            for cidx, cell in enumerate(cols):


                item = QtWidgets.QTableWidgetItem()
                item.setText(cell)
                self.table.setItem(ridx, cidx, item)
            ridx += 1



class ExpPathBrowseWidget( QtWidgets.QWidget ):
    """The SourceViewWidget info which in row 0 """

    sigOpenFile = pyqtSignal(str)

    def __init__( self, parent=None):
        QtWidgets.QWidget.__init__( self, parent )

        self.debug = False
        self.setObjectName("ExcelBrowserWidget")

        self.mainLayout = QtWidgets.QVBoxLayout()
        self.mainLayout.setSpacing(0)
        self.mainLayout.setContentsMargins(0,0,0,0)
        self.setLayout(self.mainLayout)

        self.toolbar = QtWidgets.QToolBar()
        self.mainLayout.addWidget(self.toolbar, 0)

        tbg = xwidgets.ToolBarGroup(title="Selected Dir")
        self.toolbar.addWidget(tbg)

        buttSel = xwidgets.XToolButton(text="Select")
        self.toolbar.addWidget(buttSel)

        self.txtPath = QtWidgets.QLineEdit()
        tbg.addWidget(self.txtPath)
        if G.args.dev:
            self.txtPath.setText("/home/ogt/gstl_examples/35579")

        self.tree = QtWidgets.QTreeView()
        self.mainLayout.addWidget(self.tree, 30)

        self.dirModel = QtWidgets.QFileSystemModel()
        self.dirModel.setRootPath(self.txtPath.text())
        self.dirModel.setNameFilters(["*.xlsx"])

        self.tree.setModel(self.dirModel)
        self.tree.setRootIndex( self.dirModel.index(self.txtPath.text()) )


        if G.args.dev:
            ed = os.path.join(str(self.txtPath.text()), "ATTS")
            self.tree.expand( self.dirModel.index(ed ) )

        self.tree.setColumnWidth(0, 400)

        self.tree.doubleClicked.connect(self.on_tree_double_clicked)



    def on_tree_double_clicked(self, midx):
        filename = str(self.dirModel.filePath(midx))
        self.sigOpenFile.emit(filename)




class HeaderBarWidget( QtWidgets.QWidget ):

    """Psuedo Stack as it containes a header/label and widget"""
    def __init__( self, parent=None, title=None,
                  bg=None, bg2=None, fg=None):
        QtWidgets.QWidget.__init__( self, parent )

        self.bg = bg
        self.bg2 = bg2
        self.fg = fg


        self.mainLayout = QtWidgets.QHBoxLayout()
        self.mainLayout.setContentsMargins(0,0,0,0)
        self.mainLayout.setSpacing(0)
        self.setLayout(self.mainLayout)

        self.lblLead = QtWidgets.QLabel()
        self.lblLead.setText("")
        self.lblLead.setFixedWidth(20)
        self.mainLayout.addWidget(self.lblLead, 0)

        self.lblTitle = QtWidgets.QLabel()
        self.lblTitle.setText("")
        self.mainLayout.addWidget(self.lblTitle, 10)

        self.setTitle(title)
        self.update_style()


    def setTitle(self, t):
        self.lblTitle.setText(t)

    def update_style(self):

        bg = self.bg if self.bg else "#dddddd"
        bg2 = self.bg2 if self.bg2 else "#efefef"
        fg = self.fg if self.fg else "#666666"

        sty = " color: %s; font-size: 12pt; padding: 2px 5px;" % fg
        sty += "background-color: qlineargradient(x1: 0, y1: 0, x2: 1, y2: 0, "
        sty += "stop: 0  %s" % bg
        sty += ", stop: 0.7 %s " % bg2
        sty  += "stop: 1 %s" % bg2
        sty += ");"
        self.lblLead.setStyleSheet(sty)

        sty = " color: %s; font-size: 12pt; padding: 2px 5px;"  % fg
        sty += "background-color: qlineargradient(x1: 0, y1: 0, x2: 1, y2: 0, "
        sty += "stop: 0 %s" % bg2
        sty += ", stop: 0.1 %s " % bg2
        sty  += "stop: 1 %s" % bg
        sty += ");"
        self.lblTitle.setStyleSheet(sty)

    def setCurrentIndex(self, idx):
        self.headerStack.setCurrentIndex(idx)
        self.contentStack.setCurrentIndex(idx)

    def addWidget(self, widget):
        self.mainLayout.addWidget(widget)




# -*- coding: utf-8 -*-


# Icons http://www.visual-case.it/cgi-bin/vc/GMapsIcons.pl
import os

from Qt import QtCore, QtGui, QtWidgets, QtWebEngineWidgets
from Qt import pyqtSignal, pyqtSlot
from img import Ico
import app_globals as G
from . import PROJECT_ROOT_PATH
import xwidgets

MAP_PATH = os.path.join(PROJECT_ROOT_PATH, "ogtstatic", "map")


##=================================================================================
## WebKit Sub Class
##=================================================================================

class WebViewX(QtWebEngineWidgets.QWebEnginePage):
    """Subclassed QWebView for mouse events"""

    def __init__(self, parent=None):
        QtWebEngineWidgets.QWebEnginePage.__init__(self, parent)

        self.x_mouse_point = None

    def mouseMoveEvent(self, ev):
        """Custom mouse moved
           @note: emits a "mouse_moved" event with position
        """
        # print "mouseMoveEvent", ev.globalPos()
        QtWebEngineWidgets.QWebEnginePage.mouseMoveEvent(self, ev)
        self.emit(QtCore.SIGNAL("mouse_moved"), ev.globalPos())
        self.x_mouse_point = ev.globalPos()

    def mouse_point(self):
        return self.x_mouse_point

class CL:
    loca_id = 0
    lat = 1
    lon = 2

class xMapOverviewWidget(QtWidgets.QWidget):
    debug = True

    def __init__(self, parent=None):
        QtWidgets.QWidget.__init__(self, parent)

        self.ogtDoc = None

        self.mainLayout = QtWidgets.QHBoxLayout()
        self.mainLayout.setContentsMargins(0, 0, 0, 0)
        self.mainLayout.setSpacing(0)
        self.setLayout(self.mainLayout)
        self.mainLayout.setSpacing(0)

        self.splitter = QtGui.QSplitter()
        self.mainLayout.addWidget(self.splitter)

        self.mapWidget = MapWidget()
        self.splitter.addWidget(self.mapWidget)

        self.tree = QtGui.QTreeWidget()
        self.tree.setRootIsDecorated(False)

        self.mainLayout.addWidget(self.tree)
        hi = self.tree.headerItem()
        hi.setText(CL.loca_id, "LOCA_ID")
        hi.setText(CL.lat, "Lat")
        hi.setText(CL.lon, "Lon")

        self.tree.itemDoubleClicked.connect(self.on_tree_double_clicked)

        self.splitter.setStretchFactor(0, 3)
        self.splitter.setStretchFactor(1, 1)


    def on_tree_double_clicked(self, item, cidx):
        item = self.tree.currentItem()
        if item == None:
            return

        lat = str(item.text(CL.lat))
        lon = str(item.text(CL.lon))
        self.mapWidget.zoom_to(lat=lat, lon=lon, zoom=19)

    def load_document(self, ogtDoc):

        self.ogtDoc = ogtDoc

        points =  self.ogtDoc.get_points()
        if points == None:
            return
        #print points
        for idx, p in enumerate(points):
            self.mapWidget.add_marker("xmap", id="ID_%s" % idx, lat=p['lat'], lon=p['lon'])

            item = xwidgets.XTreeWidgetItem()
            item.set(CL.loca_id, text=str(p['loca_id']), ico=Ico.TypeID, bold=True)
            #item.setIcon(CL.loca_id, Ico.icon(Ico.TypeID))
            item.setText(CL.lat, str(p['lat']))
            item.setText(CL.lon, str(p['lon']))
            self.tree.addTopLevelItem(item)

        self.mapWidget.execute_js("zoom_to_bounds()")

class MapWidget(QtWidgets.QWidget):
    debug = True

    def __init__(self, parent=None):
        QtWidgets.QWidget.__init__(self, parent)

        self.ogtDoc = None
        self.Zoom = MapZoomWidget()
        # self.gSpot = gLocation()

        self.map_initialized = False
        self.pending_markers = []
        self.locations = {}

        self.widget = None
        self.edit_mode = False
        self.dirty = None
        self.geo = None
        self.result_count = 0
        self.results = None

        ## The map local html url
        local_html_file_path = os.path.join(MAP_PATH, "map.html")
        self.map_url = QtCore.QUrl(local_html_file_path)

        #self.timer = QtCore.QTimer(self)
        #self.connect(self.timer, QtCore.SIGNAL("timeout()"), self.on_timer)

        #####################################################
        ## Popup Menu's
        #####################################################
        self.popMarker = QtGui.QMenu("pop")
        self.actionMoveMarker = self.popMarker.addAction('Move Marker', self.on_action_move_marker)

        self.popMap = QtGui.QMenu("pop")
        self.actionAddMarker = self.popMap.addAction(Ico.icon(Ico.Add), 'Save this view', self.on_action_save_view)
        self.actionAddMarker = self.popMap.addAction(Ico.icon(Ico.Add), 'Add test point', self.on_action_test_point)
        # self.actionBranch = self.popMarker.addAction(dIco.icon(dIco.Branch), 'Branch', self.on_action_branch)


        ###############
        ##  Layout
        mapLayout = QtGui.QVBoxLayout()
        mapLayout.setContentsMargins(0, 0, 0, 0)
        mapLayout.setSpacing(0)
        self.setLayout(mapLayout)
        mapLayout.setSpacing(0)

        ##======================================================
        toolbar = QtGui.QToolBar()
        mapLayout.addWidget(toolbar, 0)
        m = 4
        style = "border-bottom: 0px; border-left: 0px;"

        ## Map Type Box
        groupBoxMapType = QtGui.QGroupBox("Map Type")
        groupBoxMapType.setStyleSheet(style)
        toolbar.addWidget(groupBoxMapType)
        layoutMapTypeBox = QtGui.QHBoxLayout()
        layoutMapTypeBox.setContentsMargins(m, m, m, m)
        groupBoxMapType.setLayout(layoutMapTypeBox)

        ## Zoom Level
        groupBoxZoom = QtGui.QGroupBox("Zoom View")
        groupBoxZoom.setStyleSheet(style)
        toolbar.addWidget(groupBoxZoom)
        layoutZoomBox = QtGui.QHBoxLayout()
        layoutZoomBox.setContentsMargins(m, m, m, m)
        groupBoxZoom.setLayout(layoutZoomBox)

        #####################################################
        #### Map Type Dropdown
        #####################################################
        self.groupMapType = QtGui.QButtonGroup(self)
        self.groupMapType.setExclusive(True)
        self.connect(self.groupMapType, QtCore.SIGNAL("buttonClicked(QAbstractButton *)"), self.on_map_type)

        for b in ['RoadMap', 'Satellite', 'Hybrid', 'Terrain']:
            butt = QtGui.QToolButton()
            butt.setText(b)
            butt.setCheckable(True)
            butt.setChecked(b == "Hybrid")
            layoutMapTypeBox.addWidget(butt)
            self.groupMapType.addButton(butt)

        #########################
        ## Zoom Views
        #########################
        self.groupZoom = QtGui.QButtonGroup(self)
        self.groupZoom.setExclusive(True)
        self.connect(self.groupZoom, QtCore.SIGNAL("buttonClicked(QAbstractButton *)"), self.on_zoom_action)
        for b in self.Zoom.levels():
            butt = QtGui.QToolButton()
            butt.setText(b[0])
            butt.setProperty("zoom", QtCore.QVariant(b[1]))
            butt.setCheckable(True)
            butt.setChecked(b[0] == 'Uk')
            layoutZoomBox.addWidget(butt)
            self.groupZoom.addButton(butt)

        # tolbar3.addWidget(widgets.Widgets.ToolBarSpacer())





        ###########################################
        ## Web Browser view
        ###########################################
        self.webView = WebViewX()
        mapLayout.addWidget(self.webView, 20)
        self.webView.page().mainFrame().addToJavaScriptWindowObject("QtWidget", self)
        self.connect(self.webView, QtCore.SIGNAL("loadProgress(int)"), self.on_initialize_progress)
        self.connect(self.webView, QtCore.SIGNAL("loadFinished(bool)"), self.on_initialize_finished)

        page = self.webView.page()
        page.settings().setAttribute(QtWebKit.QWebSettings.DeveloperExtrasEnabled, True)
        page.settings().setAttribute(QtWebKit.QWebSettings.JavascriptEnabled, True)

        self.webInspector = QtWebKit.QWebInspector(self)
        self.webInspector.setPage(page)
        mapLayout.addWidget(self.webInspector)
        self.webInspector.setVisible(False)

        ###########################################
        ## Status Bar
        ###########################################
        self.statusBar = QtGui.QStatusBar()
        mapLayout.addWidget(self.statusBar)

        ## Progress
        self.progress = QtGui.QProgressBar()
        self.statusBar.addPermanentWidget(self.progress)

        ## Lat
        style = ""  # background-color: #efefef;"
        self.statusBar.addPermanentWidget(QtGui.QLabel("Lat:"))
        self.lat = QtGui.QLabel()
        self.lat.setStyleSheet(style)
        self.lat.setFixedWidth(140)
        self.lat.setAlignment(QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter)
        self.statusBar.addPermanentWidget(self.lat)

        ## Lng
        self.statusBar.addPermanentWidget(QtGui.QLabel("Lng:"))
        self.lng = QtGui.QLabel()
        self.lng.setStyleSheet(style)
        self.lng.setFixedWidth(140)
        self.lng.setAlignment(QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter)
        self.statusBar.addPermanentWidget(self.lng)

        ## Zoom
        self.statusBar.addPermanentWidget(QtGui.QLabel("Zoom:"))
        self.zoomLabel = QtGui.QLabel("")
        self.zoomLabel.setFixedWidth(20)
        self.statusBar.addPermanentWidget(self.zoomLabel)

        ###########################################
        ## Setup
        self.connect(self, QtCore.SIGNAL("zoom_changed"), self.on_zoom_changed)
        # self.connect(self, QtCore.SIGNAL("marker_right_clicked"), self.on_zoom_changed)
        self.initialize()

    ###########################################
    ## Zoom Related
    ###########################################
    def on_zoom_action(self, butt):
        zoom = butt.property("zoom").toString()
        self.set_zoom(zoom)

    def set_zoom(self, zoom):
        js_str = "set_zoom(%s)" % zoom
        self.execute_js(js_str)

    def zoom_to(self, lat=None, lon=None, zoom=None):
        if zoom == None:
            zoom = self.zoomLabel.text()
        js_str = "zoom_to(%s, %s, %s)" % (lat, lon, zoom)
        print( js_str)
        self.execute_js(js_str)

    def zoom_to_marker(self, widget, mId, zoom=None):
        if zoom == None:
            zoom = self.zoomLabel.text()
        js_str = "zoom_to_marker('%s', '%s', %s)" % (self.widget_name(widget), str(mId), str(zoom))
        self.execute_js(js_str)

    #@pyqtSlot("map_zoom_changed(QString)")
    def map_zoom_changed(self, zoom):
        self.emit(QtCore.SIGNAL("zoom_changed"), zoom)

    def on_zoom_changed(self, new_zoom_level):
        self.zoomLabel.setText(new_zoom_level)
        for a in self.groupZoom.buttons():
            checked = a.property("zoom").toString() == new_zoom_level
            a.setChecked(checked)

    ###########################################
    ## Marker Events
    ###########################################

    ## Added
    #@pyqtSlot("marker_added(QString, QString, QString)")
    def marker_added(self, widget_name, mId, idx):
        self.emit(QtCore.SIGNAL("marker_added"), str(widget_name), str(mId), str(idx))

    ## Clicked
    #@pyqtSlot("marker_clicked(QString, QString)")
    def marker_clicked(self, widget_name, myId):
        if self.debug:
            print("MV>> marker_clicked", widget_name, myId)
        self.emit(QtCore.SIGNAL("marker_clicked"), str(widget_name), str(myId))

    ## Right Clicked
    #@pyqtSlot("marker_right_clicked(QString, QString)")
    def marker_right_clicked(self, widget_name, mId):
        print(" >> marker_right_clicked", widget_name, mId)
        # self.curr_location_id = str(location_id)
        self.emit(QtCore.SIGNAL("marker_right_clicked"), str(widget_name), str(mId))

    # return
    # dic = self.locations[self.curr_location_id]
    # self.actionAccount.setText(dic['company'])
    # self.actionBranch.setText(dic['address'])
    # self.popMarker.exec_( self.webView.mouse_point() )

    ## Double Clicked
    #@pyqtSlot("marker_double_clicked(QString, QString)")
    def marker_double_clicked(self, idx, location_id):
        print("== marker_double_clicked", idx, location_id)
        self.emit(QtCore.SIGNAL("marker_double_clicked"), idx, location_id)

    ## Un-Selected
    #@pyqtSlot("marker_unselected(QString, QString)")
    def marker_unselected(self, idx, location_id):
        print("== marker_unselected", idx, location_id)
        self.emit(QtCore.SIGNAL("marker_unselected"), idx, location_id)

    ## Marker Save
    """
    @QtCore.pyqtSignature("marker_save(QString, QString, QString, QString)")
    def marker_save(self, idx, location_id, lat, lng):
        print "== marker_save", idx, location_id, lat, lng
        self.emit(QtCore.SIGNAL("marker_save"), str(idx), str(location_id), str(lat), str(lng))
        res = QtGui.QMessageBox.question(self, "Save", "Save change of position ?",  "No", "Yes")
        #print "q", res
        if res == 1:
            geo = {'lat': str(lat), 'lng': str(lng), 'method': 'manual'}
            server_vars = {'location_id': location_id, 'geo': json.dumps(geo)}
            server = dServerCall(self)
            self.connect(server, QtCore.SIGNAL('dataReady'), self.on_save_location_geo_callback)
            server.action(dAction.location_geo, server_vars, None, True)
            self.statusBar.showMessage("Saving....")	
    """

    def on_save_location_geo_callback(self, jsonData):
        #print "back", jsonData
        self.statusBar.showMessage("Postion Saved", 5000)

    #@QtCore.pyqtSignature("marker_edit(QString)")
    def marker_edit(self, idx):
        #print "== marker_edit", idx
        ##item = self.tree.findItems(idx, QtCore.Qt.MatchExactly, self.COLS.idx)[0]
        ##item.setText(self.COLS.address, "FFFF")
        ##self.tree.setCurrentItem(item)
        QtGui.QMessageBox.information(self, "ff")

    ###########################################
    ## Map Events
    ###########################################

    ## Added
    #@QtCore.pyqtSignature("map_error(QString)")
    def map_error(self, error_str):
        #print "map_error", error_str
        self.emit(QtCore.SIGNAL("map_error"), error_str)

    ## Map Initialiszed callback
    #@QtCore.pyqtSignature("set_map_init()")
    def set_map_init(self):
        self.map_initialized = True
        self.emit(QtCore.SIGNAL("map_initialized"), True)
        self.statusBar.showMessage("Map Initialised", 4000)

    #@QtCore.pyqtSignature("map_mouse_move(QString, QString)")
    def map_mouse_move(self, lat, lng):
        # print "ll", lat, lng
        self.lat.setText(lat)
        self.lng.setText(lng)

    #@QtCore.pyqtSignature("map_right_click(QString, QString)")
    def map_right_click(self, lat, lng):
        #print "RIGHT", lat, lng, self.webView.mouse_point()
        # self.emit(
        # self.map_ri
        self.geo = {'lat': lat, 'lng': lng}
        # self.lat.setText(lat)
        # self.lng.setText(lng)
        self.popMap.exec_(self.webView.mouse_point())

    # self.popMarker.exec_( self.mapFrom(self, self.pos()) )

    ###########################################
    ## Do Lookup
    ###########################################
    def do_lookup(self, widget, location_id, search):
        self.result_count = None
        self.results = []
        widget_name = widget.property("widget_name").toString()
        search = str(search)
        #print "-----------------------------------", widget
        # print ">> do_lookup >>", widget_name
        loc_id = str(location_id) if location_id else "0"
        js_str = "do_lookup('%s', '%s', '%s')" % (widget_name, loc_id, search)
        #print "javascript:", js_str
        self.execute_js(js_str)
        self.statusBar.showMessage("Looking up...")
        self.progress.setRange(0, 0)
        self.progress.show()

    ## js >> Results Count
    #@QtCore.pyqtSignature("lookup_result_count(QString, QString, QString)")
    def lookup_result_count(self, widget_name, location_id, count):
        # print "result count >>>", widget_name, location_id, count
        self.result_count = int(count)
        if self.result_count == 0:
            self.emit(QtCore.SIGNAL("lookup_results"), widget_name, None)
            self.progress.hide()
            self.statusBar.showMessage("No results found", 3000)

    ## js >> Result Item
    #@QtCore.pyqtSignature("lookup_result_item(QString, QString, QString, QString, QString)")
    def lookup_result_item(self, widget_name, location_id, address_unicode, lat, lng):
        ## Clean up unicode
        address = ""
        for l in range(0, len(address_unicode)):
            try:
                character = str(address_unicode[l])
                address = address + character
            except:
                pass
        #print ">>", address
        widget_name = str(widget_name)
        location_id = str(location_id)
        lat = str(lat)
        lng = str(lng)
        self.results.append({'location_id': location_id, 'address': address, 'lat': lat, 'lng': lng})

        ## This event fires after we got all the results from javascript - cant return dic yet
        if len(self.results) == self.result_count:
            self.emit(QtCore.SIGNAL("lookup_results"), widget_name, self.results)
            self.statusBar.showMessage("%s results found" % self.result_count, 3000)
            self.progress.hide()





    #def add_marker(self, widget, mId, lat, lng, icon_color=None, label=None):
    def add_marker(self, widget, id=None, lat=None, lon=None, icon_color=None, label=None):
        if icon_color == None:
            icon_color = "blue"
        if label == None:
            label = ""
        # print "add_marker", lat, lng, self.widget_name(widget)
        js_str = "add_marker('%s', '%s', %s, %s, '%s', '%s')" % (
            widget, id, lat, lon, icon_color, label)
        # print "js=", js_str
        self.execute_js(js_str)

    def select_marker_by_mid(self, widget, myId, emode=False):
        js_str = "select_marker_by_mid('%s', '%s', %s)" % (self.widget_name(widget), str(myId), 1 if emode else 0)
        self.execute_js(js_str)

    def set_marker_icon(self, widget, myId, icon_color):
        js_str = "set_marker_icon('%s', '%s', '%s')" % (self.widget_name(widget), str(myId), str(icon_color))
        self.execute_js(js_str)

    def clear_markers(self, widget):
        js_str = "clear_markers('%s')" % self.widget_name(widget)
        # print js_str
        self.execute_js(js_str)

    def hide_markers(self, widget, hidden):
        js_str = "hide_markers('%s', %s)" % (self.widget_name(widget), "1" if hidden else "0")
        # print js_str
        ret = self.execute_js(js_str)

    # print "ret< " , ret

    def widget_name(self, widget):
        return str(widget.property("widget_name").toString())

    ####################################################
    ## Map Control
    ####################################################
    def on_map_type(self, butt):
        js_str = "set_map_type('%s')" % str(butt.text().toLower())
        self.execute_js(js_str)

    def pan_to(self, lat, lng):
        js_str = "pan_to(%s, %s)" % (lat, lng)
        self.execute_js(js_str)

    def execute_js(self, js_str):
        #print "EXEC >", js_str
        self.webView.page().mainFrame().evaluateJavaScript(js_str)

    ###########################################
    ## Map Initialise
    ###########################################
    def initialize(self, geo=None):
        self.map_initialized = False
        self.webView.load(self.map_url)
        self.progress.setRange(0, 100)
        self.progress.setValue(0)
        self.progress.show()

    def on_initialize_progress(self, x):
        self.progress.setValue(x)

    def on_initialize_finished(self):
        self.progress.hide()

    def on_timer(self):
        if self.map_initialized:
            self.timer.stop()
            for m in self.pending_markers:
                self.add_address_marker(m[0], m[1])

    ####################################################
    ## Actions
    ####################################################
    def on_action_test_point(self):
        # print "on_action_test_point"
        self.add_marker(self.geo['lat'], self.geo['lng'])

    def on_action_move_marker(self):
        # js_str = "set_edit_mode(1)"
        self.select_marker_by_idx(self.curr_idx, True)


    def on_action_save_view(self):
        #print "on_action_save_view"
        view = self.webView.get_view();
        res = QtGui.QMessageBox.question(self, "Save", "Save Current View", "no", "Yes")
        if res == 1:
            view = self.webView.get_view();

    def get_map_view(self):
        js_str = "current_view()"
        return self.execute_js()

    ####################################################
    ## Edit Mode
    ####################################################
    def on_edit_mode(self, butt):
        #print "butt=", butt.text()
        self.edit_mode = True if butt.text() == "Edit Mode" else False
        js_str = "set_edit_mode(%s)" % ("1" if self.edit_mode else "0")
        self.execute_js(js_str)
        #print js_str
        self.buttonSave.setEnabled(self.edit_mode)
        for b in self.groupEditMode.buttons():
            if b.text() == butt.text():
                ico = dIco.Yellow
            else:
                ico = dIco.Black
            b.setIcon(dIco.icon(ico))
        #print "--- <<<"

    def on_save_button(self):
        pass
        #print "save"

    def on_map_view(self):
        #print "on_map_view"
        pass




class MapZoomWidget:
    """Zoom Helper Class.
      - Helps with map level zoom handling
      - Iniial zoom levels are Roof, street, Local, Area, Uk

    @todo: make zoom levels constants
    """

    def __init__(self):

        self.i_levels = {}
        self.labels = {}
        for level in [[19, 'Roof'], [17, 'Street'], [14, 'Local'], [8, 'Area'], [6, 'Uk']]:
            self.add_level(level[0], level[1])

    def add_level(self, level, label):
        self.i_levels[label] = level
        self.labels[level] = label

    def level(self, label=None):
        if label == None:
            label = 'Local'
        return self.i_levels[label]

    def levels(self):
        levels = self.labels.keys()
        levels.sort(reverse=True)
        ret_list = []
        for level in levels:
            ret_list.append([self.labels[level], level])
        return ret_list

    def make_menu(self, parent, menu, callback):
        zoomMenu = menu.addMenu("Zoom to")
        groupZoomLevels = QtGui.QActionGroup(parent)
        parent.connect(groupZoomLevels, QtCore.SIGNAL("selected(QAction *)"), callback)  # self.on_zoom_action)
        for z in self.levels():
            act = zoomMenu.addAction(z[0])
            act.setProperty("zoom", QtCore.QVariant(z[1]))
            groupZoomLevels.addAction(act)






"""
Converts british national grid to lat lon
Author: Hannah Fry
http://www.hannahfry.co.uk/blog/2012/02/01/converting-british-national-grid-to-latitude-and-longitude-ii
https://github.com/fmalina/bng_latlon
"""

from math import sqrt, pi, sin, cos, tan, atan2


def OSGB36toWGS84(E, N):
    """ Accept The Ordnance Survey National Grid eastings and northings.
    Return latitude and longitude coordinates.
    Usage:
    >>> from bng_to_latlon import OSGB36toWGS84
    >>> OSGB36toWGS84(538890, 177320)
    (51.47779538331092, -0.0014016837826672265)
    >>> OSGB36toWGS84(352500.2, 401400)
    (53.507129843104195, -2.7176599627343263)
    """
    # The Airy 1830 semi-major and semi-minor axes used for OSGB36 (m)
    a, b = 6377563.396, 6356256.909
    F0 = 0.9996012717  # scale factor on the central meridian

    # Latitude and longtitude of true origin (radians)
    lat0 = 49*pi/180
    lon0 = -2*pi/180  # longtitude of central meridian

    # Northing & easting of true origin (m)
    N0, E0 = -100000, 400000
    e2 = 1 - (b*b)/(a*a)  # eccentricity squared
    n = (a-b)/(a+b)

    # Initialise the iterative variables
    lat, M = lat0, 0

    while N-N0-M >= 0.00001:  # Accurate to 0.01mm
        lat = (N-N0-M)/(a*F0) + lat
        M1 = (1 + n + (5./4)*n**2 + (5./4)*n**3) * (lat-lat0)
        M2 = (3*n + 3*n**2 + (21./8)*n**3) * sin(lat-lat0) * cos(lat+lat0)
        M3 = ((15./8)*n**2 + (15./8)*n**3) * sin(2*(lat-lat0)) * cos(2*(lat+lat0))
        M4 = (35./24)*n**3 * sin(3*(lat-lat0)) * cos(3*(lat+lat0))
        # meridional arc
        M = b * F0 * (M1 - M2 + M3 - M4)

    # transverse radius of curvature
    nu = a*F0/sqrt(1-e2*sin(lat)**2)

    # meridional radius of curvature
    rho = a*F0*(1-e2)*(1-e2*sin(lat)**2)**(-1.5)
    eta2 = nu/rho-1

    sec_lat = 1./cos(lat)
    VII = tan(lat)/(2*rho*nu)
    VIII = tan(lat)/(24*rho*nu**3)*(5+3*tan(lat)**2+eta2-9*tan(lat)**2*eta2)
    IX = tan(lat)/(720*rho*nu**5)*(61+90*tan(lat)**2+45*tan(lat)**4)
    X = sec_lat/nu
    XI = sec_lat/(6*nu**3)*(nu/rho+2*tan(lat)**2)
    XII = sec_lat/(120*nu**5)*(5+28*tan(lat)**2+24*tan(lat)**4)
    XIIA = sec_lat/(5040*nu**7)*(61+662*tan(lat)**2+1320*tan(lat)**4+720*tan(lat)**6)
    dE = E-E0

    # These are on the wrong ellipsoid currently: Airy 1830 (denoted by _1)
    lat_1 = lat - VII*dE**2 + VIII*dE**4 - IX*dE**6
    lon_1 = lon0 + X*dE - XI*dE**3 + XII*dE**5 - XIIA*dE**7

    # Want to convert to the GRS80 ellipsoid.
    # First convert to cartesian from spherical polar coordinates
    H = 0  # Third spherical coord.
    x_1 = (nu/F0 + H)*cos(lat_1)*cos(lon_1)
    y_1 = (nu/F0 + H)*cos(lat_1)*sin(lon_1)
    z_1 = ((1-e2)*nu/F0 + H)*sin(lat_1)

    # Perform Helmut transform (to go between Airy 1830 (_1) and GRS80 (_2))
    s = -20.4894*10**-6  # The scale factor -1
    # The translations along x, y, z axes respectively
    tx, ty, tz = 446.448, -125.157, + 542.060
    # The rotations along x, y, z respectively (in seconds)
    rxs, rys, rzs = 0.1502, 0.2470, 0.8421

    # convert seconds to radians
    def sec_to_rad(x): return x*pi/(180*3600.)

    rx, ry, rz = [sec_to_rad(x) for x in (rxs, rys, rzs)]  # (in radians)
    x_2 = tx + (1+s)*x_1 + (-rz)*y_1 + (ry)*z_1
    y_2 = ty + (rz)*x_1 + (1+s)*y_1 + (-rx)*z_1
    z_2 = tz + (-ry)*x_1 + (rx)*y_1 + (1+s)*z_1

    # Back to spherical polar coordinates from cartesian
    # Need some of the characteristics of the new ellipsoid

    # The GSR80 semi-major and semi-minor axes used for WGS84(m)
    a_2, b_2 = 6378137.000, 6356752.3141
    e2_2 = 1 - (b_2*b_2)/(a_2*a_2)  # The eccentricity of the GRS80 ellipsoid
    p = sqrt(x_2**2 + y_2**2)

    # Lat is obtained by an iterative proceedure:
    lat = atan2(z_2, (p*(1-e2_2)))  # Initial value
    latold = 2*pi
    while abs(lat - latold) > 10**-16:
        lat, latold = latold, lat
        nu_2 = a_2/sqrt(1-e2_2*sin(latold)**2)
        lat = atan2(z_2+e2_2*nu_2*sin(latold), p)

    # Lon and height are then pretty easy
    lon = atan2(y_2, x_2)
    H = p/cos(lat) - nu_2

    # Uncomment this line if you want to print the results
    # print([(lat-lat_1)*180/pi, (lon - lon_1)*180/pi])

    # Convert to degrees
    lat = lat*180/pi
    lon = lon*180/pi

    # Job's a good'n.
    return lat, lon


if __name__ == "__main__":
    import doctest
    doctest.testmod()

    import csv

    # Read in from a file
    BNG = csv.reader(open('BNG.csv', 'rU'), delimiter=',')
    next(BNG)

    # Get the output file ready
    #changes made because of: #https://stackoverflow.com/questions/3348460/csv-file-written-with-python-has-blank-lines-between-each-row
    output_file = open('BNGandLatLon.csv', 'w+', newline='')
    output = csv.writer(output_file, delimiter=',')
    output.writerow(['Lat', 'Lon', 'E', 'N'])

    # Loop through the data
    for E, N in BNG:
        lat, lon = OSGB36toWGS84(float(E), float(N))
        output.writerow([str(lat), str(lon), str(E), str(N)])
    # Close the output file
    output_file.close()



# -*- coding: utf-8 -*-


import os
import sys

HERE_PATH =  os.path.abspath( os.path.dirname( __file__))
# make sure ogtgui in syspath
if not HERE_PATH in sys.path:
    sys.path.insert(0, HERE_PATH)

PROJECT_ROOT_PATH = os.path.abspath(os.path.join(HERE_PATH, ".."))
"""Root dir of this project"""

ICONS_PATH = os.path.join(PROJECT_ROOT_PATH, "ogtstatic", "icons")

DEFAULT_ROW_HEIGHT = 25



#!/usr/bin/env python3
# coding: utf-8
""""""
from ogt.ogt_validate import OGTError
from ogt import CELL_COLORS


class OGTCell:

    def __init__(self, parent_heading=None, parent_group=None, src_ref=None, lidx=None, cidx=None,
                 raw=None):
        """Core class that contains a cell value

        :param parent_heading: the parent heading this cell belong to
        :type parent_heading: OGTHeading
        :param src_ref: source reference
        :type src_ref: str
        :param lidx: line index in source
        :type lidx: int
        :param cidx: column index in source
        :type cidx: int
        :param raw: the raw value
        :type raw: str

        """

        self.parent_group = parent_group

        self.parent_heading = parent_heading
        """The heading this cell is within

        :type: OGTHeading

        """
        self.src_ref = src_ref
        """Source reference

        :type: str
        """

        self.lidx = lidx
        """Line index within source

        :type: int
        """

        self.cidx = cidx
        """column index

        :type: int
        """

        self.raw = raw
        """Raw value

        :type: str
        """

        self.value = None
        """actual value

        :type: str or None
        """

        self._import_warnings = []
        self._warnings = []
        self._errors = []
        self._strip()

    def __repr__(self):
        return "<OGTCell `%s`>" % self.value

    def _strip(self):
        # check and remove whitespace, called at import
        if self.raw is None:
            return
        self.value = self.raw.strip()
        if self.value != self.raw:

            if self.value == self.raw.lstrip():
                error = OGTError(self, "Leading white space `%s`" % self.raw, warn=True)
                self._import_warnings.append(error)

            elif self.value == self.raw.rstrip():
                error = OGTError(self, "Trailing white space `%s`" % self.raw, warn=True)
                self._import_warnings.append(error)

            else:
                error = OGTError(self, "White space`%s`" % self.raw, warn=True)
                self._import_warnings.append(error)

    def clear_errors(self):
        """Clears all errors and warnings"""
        del self._import_warnings[:]
        del self._warnings[:]
        del self._errors[:]

    @property
    def warnings_count(self):
        """"""
        return len(self._import_warnings) + len(self._warnings)

    def warnings_list(self):
        """"""
        return self._import_warnings + self._warnings

    def errors_list(self):
        """"""
        return self._errors

    @property
    def errors_count(self):
        """"""
        return len(self._errors)

    def add_error(self, err):
        """
        :param err: error/warning instance to add
        :type err: OGTError
        """
        if err.err_type == OGTError.WARN:
            self._warnings.append(err)
        else:
            self._errors.append(err)

    def background(self):
        """Background color

        Not really logic, but the idea is to color cells in code, and across www, desktop, mobile

        See :data:`~ogt.__init__.CELL_COLORS`
        """
        if self.errors_count:
            return CELL_COLORS.err_bg

        if self.warnings_count:
            return CELL_COLORS.warn_bg

        return CELL_COLORS.ok_bg



import os
import tempfile

####################################################
# Note this file is for variables only, no imports
# except standard libs where necessary
# as its imported for docs gen etc
####################################################

PROJECT_VERSION = "0.0.1"

PROJECT_SHORT = "ogt-py"
PROJECT_LONG = "ogt-py"

PROJECT_DESCRIPTION = "Python code that eats geotechcical data"
PROJECT_CONTACT = "ogt@daffodil.uk.com"

PROJECT_DOMAIN = "open-geotechnical.gitlab.io"
PROJECT_WWW = "http://open-geotechnical.gitlab.io/"
PROJECT_HOME = "https://gitlab.com/open-geotechnical/ogt-py"
PROJECT_ISSUES = "https://gitlab.com/open-geotechnical/ogt-py/issues"
PROJECT_API_DOCS = "http://open-geotechnical.gitlab.io/ogt-py"


def get_project_info():
    """
    :return: A `dict` with the project info
    """
    return dict(
        version=PROJECT_VERSION,
        short=PROJECT_SHORT,
        long=PROJECT_LONG,
        description=PROJECT_DESCRIPTION,
        contact=PROJECT_CONTACT,
        domain=PROJECT_DOMAIN,
        www=PROJECT_WWW,
        home=PROJECT_HOME,
        issues=PROJECT_ISSUES,
        api_docs=PROJECT_API_DOCS
    )


HERE_PATH = os.path.abspath(os.path.dirname(__file__))

PROJECT_ROOT_PATH = os.path.abspath(os.path.join(HERE_PATH, ".."))
"""Root dir of the project"""

TEMP_DIR = tempfile.gettempdir()
TEMP_WORKSPACE = os.path.join(TEMP_DIR, "temp_workspace")
"""Path to temporary directory"""

EXAMPLES_DIR = os.path.join(TEMP_DIR, "example_files")
"""Path to examples folder"""

USER_HOME = os.path.expanduser("~")
"""Path to users home dir"""

USER_TEMP = os.path.join(USER_HOME, "ogt-workspace")
"""Path to open-getechnical cache directory"""

FORMATS = ["json", "geojson", "yaml", "xlsx", "ags4", "ags"]
"""Formats allowed, depending on libs installed"""


class CELL_COLORS:
    empty_bg = "#666666"
    ok_bg = "#E7FFDC"
    err_bg = "#FFC5C5"
    warn_bg = "#F7F89F"


class COLOR:
    required = "purple"



# -*- coding: utf-8 -*-


import os
import datetime

from Qt import QtGui, QtWidgets, QtCore, Qt, pyqtSignal

import openpyxl  # import load_workbook
import openpyxl.utils

from ogt import ags4
import xwidgets
import ogtgui_widgets






class ExcelWorkbookWidget(QtWidgets.QWidget):

    def __init__(self, parent=None, filename=None):
        QtWidgets.QWidget.__init__(self, parent)

        self.debug = False
        self.filename = filename
        self.setObjectName("ExcelWorkbookWidget")

        self.mainLayout = xwidgets.vlayout()
        self.setLayout(self.mainLayout)


        self.tabWidget = QtWidgets.QTabWidget()
        self.mainLayout.addWidget(self.tabWidget)



    def load_workbook(self, filename=None):


        progress = QtWidgets.QProgressDialog(self)
        progress.setWindowModality(Qt.WindowModal)
        progress.setLabelText("Loading...\n%s" % filename)
        progress.setRange(0, 0)
        progress.setValue(0)
        progress.forceShow()

        #filename = str(filename)
        #print "==", self.filename, type(self.filename), self
        wb = openpyxl.load_workbook(filename=self.filename, data_only=True)
        for sheet in wb.worksheets:
            sheetWidget = ExcelSheetWidget(sheet=sheet)
            self.tabWidget.addTab(sheetWidget, sheet.title)

        progress.hide()



class ExcelSheetWidget(QtWidgets.QWidget):

    def __init__(self, parent=None, sheet=None):
        QtWidgets.QWidget.__init__(self, parent)


        self.setObjectName("ExcelSheetWidget")

        self.sheet = sheet

        self.mainLayout = xwidgets.vlayout()
        self.setLayout(self.mainLayout)

        self.table = QtWidgets.QTableWidget()
        self.mainLayout.addWidget(self.table)

        self.load_sheet()

    def load_sheet(self):
        #print "load_sheet", self.sheet, self
        #print self.sheet['A1'].value, self.sheet['A2']

        last_col = None

        maxi = 100
        for ridx in range(0, maxi):
            if self.table.rowCount() < ridx + 1:
                self.table.setRowCount(ridx + 1)
            for cidx in range(0, maxi):
                if self.table.columnCount() < cidx + 1:
                   self.table.setColumnCount(cidx + 1)
                   item = QtWidgets.QTableWidgetItem()
                   item.setText(openpyxl.utils.get_column_letter(cidx + 1))
                   self.table.setHorizontalHeaderItem(cidx, item)
                #cell_ref = ogt_excel.rowcol_to_cell(ridx, cidx)
                cell_ref = openpyxl.utils.get_column_letter(cidx + 1) + str(ridx + 1)
                #print("cell=", cell_ref)
                v = self.sheet[cell_ref].value
                #print ridx, cidx, cell_ref, v
                if v != None and isinstance(v, str) and v.upper() == "STOP":
                    last_col = cidx
                    #dsa
                else:
                    if v == None:
                        cv = ""
                    #elif isinstance(v, long):
                    #    cv = str(v)

                    elif isinstance(v, float):
                        cv = str(v)
                    elif isinstance(v, datetime.datetime):
                        cv = str(v)
                    else:
                        cv = v
                    item = QtWidgets.QTableWidgetItem()
                    item.setText( str(cv))
                    self.table.setItem(ridx, cidx, item)
                    #print(item, cv)

                    if ridx == 0 and v != None:
                        #print("v=", v)
                        if v.upper() == v:
                            grp = ags4.DD.heading(v)
                            #print("grp=", grp)
                            if grp:
                                #
                                #print(YES)
                                item.setBackground(QtGui.QColor("#C4FAA9"))



class ExcelBrowserWidget( QtWidgets.QWidget ):
    """T"""

    def __init__( self, parent=None, file_path=None, path_browser=False):
        super().__init__( parent )

        self.debug = False
        self.setObjectName("ExcelBrowserWidget")

        self.mainLayout = xwidgets.vlayout()
        self.setLayout(self.mainLayout)

        self.splitter = QtWidgets.QSplitter()
        self.mainLayout.addWidget(self.splitter, 10)

        self.tabWidget = QtWidgets.QTabWidget()
        self.splitter.addWidget(self.tabWidget)

        if path_browser:
            self.pathBrowse = ogtgui_widgets.ExpPathBrowseWidget()
            self.splitter.addWidget(self.pathBrowse)
            self.pathBrowse.sigOpenFile.connect(self.on_open_excel)

        self.splitter.setStretchFactor(0, 4)
        if path_browser:
            self.splitter.setStretchFactor(1, 1)




        if file_path:
            self.on_open_excel(file_path)

    def init_load(self):
        pass


    def on_open_excel(self, file_path):
        """
        open excel file
        """
        filename = str(file_path)

        widget = ExcelWorkbookWidget(filename=filename)
        self.tabWidget.addTab(widget, os.path.basename(filename))
        widget.load_workbook()
        #progress.hide()

class ExcelBrowserDialog( QtWidgets.QDialog ):
    """T"""

    def __init__( self, parent=None, file_path=None):
        super().__init__( parent )

        #self.file_path = file_path
        assert file_path != None

        self.mainLayout = xwidgets.vlayout()
        self.setLayout(self.mainLayout)

        self.excelBrowser = ExcelBrowserWidget(self, file_path=file_path)
        self.mainLayout.addWidget(self.excelBrowser)


        self.setWindowState(Qt.WindowMaximized)



#!/usr/bin/env python3
# coding: utf-8

from ogt.ags4 import DD


class OGTHeading:
    """"""
    def __init__(self, parent_group, hcell):

        self.parent_group = parent_group
        """parent group this heading belongs to

        :type: OGTGroup
        """

        self.head_cell = hcell
        """The HEADING cell

        :type: OGTCell
        """

        self.unit_cell = None
        """The UNIT cell

        :type: OGTCell
        """

        self.data_type_cell = None
        """The TYPE cell

        :type: OGTCell
        """

    def emit(self):
        """"""
        self.parent_group.emit()

    def __repr__(self):
        return "<Heading `%s`>" % self.head_code

    @property
    def head_code(self):
        """"""
        return self.head_cell.value

    @property
    def head_description(self):
        """"""
        data_dict = self.parent_group.data_dict()
        if data_dict and self.head_code in data_dict['headings_index']:
            idx = data_dict['headings_index'].index(self.head_code)
            return data_dict['headings'][idx]["head_description"]

        return "?"

    @property
    def required(self):
        """"""
        data_dict = self.parent_group.data_dict()
        if data_dict and self.head_code in data_dict['headings']:
            return data_dict['headings'][self.head_code]["required"]

        return None

    @property
    def key(self):
        """"""
        data_dict = self.parent_group.data_dict()
        if data_dict and self.head_code in data_dict['headings']:
            return data_dict['headings'][self.head_code]["key"]

        return None

    def set_unit(self, cell):
        """"""
        # TODO Check valid
        # dscell.value, errs = ags4.validate_clean_str(cell.value, upper=True)
        # self.ogtGroup.add_errors(errs)
        self.unit_cell = cell

    @property
    def unit(self):
        """"""
        return self.unit_cell.value

    @property
    def unit_label(self):
        """"""
        if self.unit:
            return self.unit
        return ""

    def set_data_type(self, cell):
        """"""
        self.data_type_cell = cell

    @property
    def data_type(self):
        """"""
        return self.data_type_cell.value

    @property
    def type_label(self):
        """"""
        if self.data_type:
            return self.data_type
        return ""

    def errors_count(self):
        """"""
        error_count = 0
        for cell in [self.head_cell, self.unit_cell, self.data_type_cell]:
            if cell is not None:
                error_count += cell.errors_count()

        return error_count

    def abbrs_picklist(self, ags=True, custom=True):
        """The piclist is ags and custom to this document

        :return: Abbrs for this heading
        """
        if self.data_type != "PA":
            return None

        return self.parent_group.parentDoc.abbrs(self.head_code)
        # FIXME is any of the code after this line used?
        drecs = {}
        if custom:
            grp = self.parent_group.parentDoc.group_from_code("ABBR")
            if grp:
                for cells_dic in grp.data:
                    cell_h = cells_dic.get("ABBR_HDNG")
                    if cell_h and cell_h.value == self.head_code:
                        dic = dict(code=cells_dic.get("ABBR_CODE").value,
                                   description=cells_dic.get("ABBR_DESC").value,
                                   list=cells_dic.get("ABBR_LIST").value,
                                   custom=True)
                        drecs[cells_dic.get("ABBR_CODE").value] = dic

        if ags:
            arecs = list(DD.abbrs_dict.get(self.head_code))
            for rec in arecs:
                drecs[rec['code']] = rec

        return [drecs[ki] for ki in sorted(drecs)]



# -*- coding: utf-8 -*-
"""
@author: Peter Morgan <pete@daffodil.uk.com>
"""

import json

from Qt import Qt, QtCore


#import app_globals as G

class Srv(object):

    def __init__(self, name=None, ki=None, url=None):
        self.name = name
        self.ki = ki
        self.url = url

    def __repr__(self):
        return "<Src url=%s>"

class _XSettingsImp( QtCore.QObject ):

    CURR_SERVER_KI = "current/server_ki"

    def __init__( self, parent=None ):
        QtCore.QObject.__init__( self, parent )

        self.qsettings = QtCore.QSettings("AGS", "ags-desktop-pyqt")

        self.servers = {}

        self.servers['online'] = Srv(**{'name': 'Online', "ki": "live",
                                    'url': 'http://ags.daffodil.uk.com'})

        self.servers['dev'] = Srv(**{'name': 'Developer', "ki": "dev",
                                    'url': 'http://localhost:13777'})

    def set_current_server( self, ki ):
        self.qsettings.setValue( self.CURR_SERVER_KI, ki )
        self.qsettings.sync()
        if ki in self.servers:
            self._current_server = self.servers[ki]
        return self._current_server

    def current_server( self ):
        ki = str( self.qsettings.value( self.CURR_SERVER_KI ) )
        if ki in self.servers:
            return self.server[ki]
        return self.servers["dev"]

    ##==============================
    ## Window Save/Restore
    def save_window( self, window ):
        name = window.objectName()
        if len(name) == 0:
            return
        self.qsettings.setValue( "window/%s/geometry" % name, QtCore.QVariant( window.saveGeometry() ) )

    def restore_window( self, window ):
        name = str(window.objectName())
        if len(name) == 0:
            return
        window.restoreGeometry( self.qsettings.value( "window/%s/geometry" % name ) )


    def save_splitter(self, splitter):
        wname = str(splitter.objectName())
        if not wname:
            print ("Splitter has no name", splitter)
        self.qsettings.setValue("splitter/%s" % wname, splitter.saveState())

    def restore_splitter(self, splitter):
        wname = str(splitter.objectName())
        if not wname :
            print ("Splitter has no name")
            return
        v = self.qsettings.value("splitter/%s" % wname)
        if v:
            splitter.restoreState(v)

    def save_list(self, key, lst):
        self.qsettings.setValue(key, json.dumps(lst))
        self.qsettings.sync()

    def get_list(self, key):
        s = self.qsettings.value(key)
        if s == None or s == "":
            return []
        return json.loads(s)


Settings = _XSettingsImp()



"""
Converts lat lon (WGS84) to british national grid (OSBG36)
Author: Hannah Fry
http://www.hannahfry.co.uk/blog/2012/02/01/converting-latitude-and-longitude-to-british-national-grid
"""
from math import sqrt, pi, sin, cos, tan, atan2


def WGS84toOSGB36(lat, lon):
    """ Accept latitude and longitude as used in GPS.
    Return OSGB grid coordinates: eastings and northings.

    Usage:
    >>> from latlon_to_bng import WGS84toOSGB36
    >>> WGS84toOSGB36(51.4778, -0.0014)
    (538890.1053365842, 177320.49650700082)
    >>> WGS84toOSGB36(53.50713, -2.71766)
    (352500.19520169357, 401400.01483428996)
    """
    # First convert to radians
    # These are on the wrong ellipsoid currently: GRS80. (Denoted by _1)
    lat_1 = lat*pi/180
    lon_1 = lon*pi/180

    # Want to convert to the Airy 1830 ellipsoid, which has the following:
    # The GSR80 semi-major and semi-minor axes used for WGS84(m)
    a_1, b_1 = 6378137.000, 6356752.3141
    e2_1 = 1 - (b_1*b_1)/(a_1*a_1)  # The eccentricity of the GRS80 ellipsoid
    nu_1 = a_1/sqrt(1-e2_1*sin(lat_1)**2)

    # First convert to cartesian from spherical polar coordinates
    H = 0  # Third spherical coord.
    x_1 = (nu_1 + H)*cos(lat_1)*cos(lon_1)
    y_1 = (nu_1 + H)*cos(lat_1)*sin(lon_1)
    z_1 = ((1-e2_1)*nu_1 + H)*sin(lat_1)

    # Perform Helmut transform (to go between GRS80 (_1) and Airy 1830 (_2))
    s = 20.4894*10**-6  # The scale factor -1
    # The translations along x,y,z axes respectively
    tx, ty, tz = -446.448, 125.157, -542.060
    # The rotations along x,y,z respectively, in seconds
    rxs, rys, rzs = -0.1502, -0.2470, -0.8421
    # In radians
    rx, ry, rz = rxs*pi/(180*3600.), rys*pi/(180*3600.), rzs*pi/(180*3600.)
    x_2 = tx + (1+s)*x_1 + (-rz)*y_1 + (ry)*z_1
    y_2 = ty + (rz)*x_1 + (1+s)*y_1 + (-rx)*z_1
    z_2 = tz + (-ry)*x_1 + (rx)*y_1 + (1+s)*z_1

    # Back to spherical polar coordinates from cartesian
    # Need some of the characteristics of the new ellipsoid
    # The GSR80 semi-major and semi-minor axes used for WGS84(m)
    a, b = 6377563.396, 6356256.909
    e2 = 1 - (b*b)/(a*a)  # The eccentricity of the Airy 1830 ellipsoid
    p = sqrt(x_2**2 + y_2**2)

    # Lat is obtained by an iterative proceedure:
    lat = atan2(z_2, (p*(1-e2)))  # Initial value
    latold = 2*pi
    while abs(lat - latold) > 10**-16:
        lat, latold = latold, lat
        nu = a/sqrt(1-e2*sin(latold)**2)
        lat = atan2(z_2+e2*nu*sin(latold), p)

    # Lon and height are then pretty easy
    lon = atan2(y_2, x_2)
    H = p/cos(lat) - nu

    # E, N are the British national grid coordinates - eastings and northings
    F0 = 0.9996012717  # scale factor on the central meridian
    lat0 = 49*pi/180  # Latitude of true origin (radians)
    lon0 = -2*pi/180  # Longtitude of true origin and central meridian (radians)
    N0, E0 = -100000, 400000  # Northing & easting of true origin (m)
    n = (a-b)/(a+b)

    # meridional radius of curvature
    rho = a*F0*(1-e2)*(1-e2*sin(lat)**2)**(-1.5)
    eta2 = nu*F0/rho-1

    M1 = (1 + n + (5/4)*n**2 + (5/4)*n**3) * (lat-lat0)
    M2 = (3*n + 3*n**2 + (21/8)*n**3) * sin(lat-lat0) * cos(lat+lat0)
    M3 = ((15/8)*n**2 + (15/8)*n**3) * sin(2*(lat-lat0)) * cos(2*(lat+lat0))
    M4 = (35/24)*n**3 * sin(3*(lat-lat0)) * cos(3*(lat+lat0))

    # meridional arc
    M = b * F0 * (M1 - M2 + M3 - M4)

    I = M + N0
    II = nu*F0*sin(lat)*cos(lat)/2
    III = nu*F0*sin(lat)*cos(lat)**3*(5 - tan(lat)**2 + 9*eta2)/24
    IIIA = nu*F0*sin(lat)*cos(lat)**5*(61 - 58*tan(lat)**2 + tan(lat)**4)/720
    IV = nu*F0*cos(lat)
    V = nu*F0*cos(lat)**3*(nu/rho - tan(lat)**2)/6
    VI = nu*F0*cos(lat)**5*(5 - 18*tan(lat)**2 + tan(lat)**4 + 14*eta2 - 58*eta2*tan(lat)**2)/120

    N = I + II*(lon-lon0)**2 + III*(lon-lon0)**4 + IIIA*(lon-lon0)**6
    E = E0 + IV*(lon-lon0) + V*(lon-lon0)**3 + VI*(lon-lon0)**5

    # Job's a good'n.
    return E, N


if __name__ == "__main__":
    import doctest
    doctest.testmod()

    import csv

    # Read in from a file
    lat_lon = csv.reader(open('LatLon.csv', 'rU'), delimiter=',')
    next(lat_lon)

    # Get the output file ready
    # Issue encountered: https://stackoverflow.com/questions/3348460/csv-file-written-with-python-has-blank-lines-between-each-row
    output_file = open('LatLonandBNG.csv', 'w+', newline='')
    output = csv.writer(output_file, delimiter=',')
    output.writerow(['Lat', 'Lon', 'E', 'N'])

    # Loop through the data
    for line in lat_lon:
        lat = line[0]
        lon = line[1]
        E, N = WGS84toOSGB36(float(lat), float(lon))
        output.writerow([str(lat), str(lon), str(E), str(N)])

    # Close the output file
    output_file.close()



# -*- coding: utf-8 -*-

import os
import sys

from ogt import ags4

from Qt import Qt, QtGui, QtCore, QtWidgets

import app_globals as G
from settings import Settings

import ags4_widgets
import ags4_models
import www_client
import ogtgui_widgets
import ogtgui_project
import ogtgui_projects
import ogtgui_excel

from img import Ico
import xwidgets

APP_TITLE = "ogt-py"


class MainWindow(QtWidgets.QMainWindow):

    def on_after(self):
        """This is called after __init__, a few moment later"""

        # self.on_browse_ags4()


        if G.args.dev:
            # fn = "/home/ogt/AGS4-example-wrd.ags"
            # fn = "/home/ogt/ags-play/example_files/pete_stuff/simple.ags"
            # fn = "/home/geo2lab/z_drive/jobs/40101-40200/40153/AGS/PT 1.ags"
            fn = "/home/ogt/ags-play/example_files/pete_stuff/example_schedule.ags"
            #fn = "/home/ogt/ogt-py/etc/simple.ags"
            #self.load_ags4_file(fn)
            pass

    def on_open_excel(self):

        filePath, _filter = QtWidgets.QFileDialog.getOpenFileName(self, "Select Excel", "", "Excel(*.xlsx)")
        if filePath:
            self.load_excel(filePath)

    def load_excel(self, file_path):

            w = ogtgui_excel.ExcelBrowserDialog(parent=self, file_path=file_path)
            w.setWindowState(Qt.WindowMaximized)
            w.exec_()


    def __init__( self, args ):
        super().__init__()

        # ===========================================
        # Bit hacky, but sticking stuff in Globals :ref:`app_globals`
        # and initialising stuff
        G.mainWindow = self
        G.args = args # command args

        #G.settings = settings.XSettings(self)

        self.server = www_client.ServerConnection( self )
        self.server.response.connect( self.on_www_request_finished )
        G.server = self.server

        G.ags = ags4_models.Ags4Object()


        ##===============================================
        # Main window stuff
        self.setObjectName("OGTMainWindow")
        QtWidgets.QApplication.setStyle( QtWidgets.QStyleFactory.create( 'Cleanlooks' ) )
        self.setWindowTitle("%s - %s" % (APP_TITLE,G.version))
        self.setWindowIcon(Ico.icon(Ico.FavIcon))


        ##=================================================
        ## Menus - or a welsh meniw ;-)

        #=======
        ## File
        self.menuFile = self.menuBar().addMenu("File")

        self.actionOpen = self.menuFile.addAction("Open Ags4 File", self.on_open_ags_file)
        self.actionRecent = self.menuFile.addAction("Recent")
        self.actionRecent.setMenu(QtWidgets.QMenu())
        self.actionRecent.menu().triggered.connect(self.on_open_recent)
        self.menuFile.addSeparator()
        self.actionOpenExcel = self.menuFile.addAction("Open Excel", self.on_open_excel)


        self.actionQuit = self.menuFile.addAction(Ico.icon(Ico.Quit), "Quit", self.on_quit)

        #=======
        ## Projects
        self.menuProjects = self.menuBar().addMenu("Projects")

        self.widgetProjects = QtWidgets.QWidgetAction(self.menuProjects)
        self.projectsWidget = ogtgui_projects.OgtProjectsWidget(self)
        self.widgetProjects.setDefaultWidget(self.projectsWidget)

        self.actionProjects = self.menuProjects.addAction(self.widgetProjects)

        self.actionNewProject = self.menuProjects.addAction(Ico.icon(Ico.Add), "New Project", self.on_new_project)
        self.menuProjects.addSeparator()

        #=======
        ## View
        self.menuViews = self.menuBar().addMenu("View")
        self.actionAgs4Browse = self.menuViews.addAction(Ico.icon(Ico.Ags4), "AGS4", self.on_browse_ags4)
        self.actionAgs4Browse.setCheckable(True)

        #=======
        ## Examples - its an example widget within
        self.menuExamples = self.menuBar().addMenu("Examples")

        self.widgetActionExamples = QtWidgets.QWidgetAction(self.menuExamples)
        self.examplesWidget = ogtgui_widgets.ExamplesWidget(self)
        self.examplesWidget.setMinimumHeight(600)
        self.widgetActionExamples.setDefaultWidget(self.examplesWidget)
        self.examplesWidget.sigFileSelected.connect(self.load_ags4_example)

        self.actionExamples = self.menuExamples.addAction(self.widgetActionExamples)

        # help meniw (meniw = woman in welsh, eg a cafe.. U want the menu?  no food first ;-))
        self.menuHelp = self.menuBar().addMenu("Help")

        self.menuHelp.addAction("ogt-py")

        ##===========================
        ## Top Bar
        self.toolBar = QtWidgets.QToolBar()
        self.toolBar.setContentsMargins(0, 0, 0, 0)
        self.toolBar.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
        self.toolBar.setMovable(False)
        self.addToolBar(Qt.TopToolBarArea, self.toolBar)

        self.toolBar.addAction(self.actionAgs4Browse)

        self.toolBar.addSeparator()

        ### add a Banner and logo
        logoWidget = QtWidgets.QWidget()
        lwLay = xwidgets.hlayout(margin=0)
        logoWidget.setLayout(lwLay)
        self.toolBar.addWidget(logoWidget)

        self.lblBanner = QtWidgets.QLabel()
        self.lblBanner.setText(APP_TITLE)
        self.lblBanner.setAlignment(Qt.AlignRight|Qt.AlignVCenter)
        self.lblBanner.setSizePolicy(QtWidgets.QSizePolicy.Expanding,QtWidgets.QSizePolicy.Minimum)

        final_color = "#9E5015"
        sty = "font-style:italic; font-weight: bold;  color: #dddddd; margin: 0; font-size: 12pt; font-family: arial;"
        sty += "padding: 2px;"
        if True:
            sty += "background-color: qlineargradient(x1: 0, y1: 0, x2: 1, y2: 0, "
            sty += "stop: 0 transparent "
            sty += ", stop: 0.1 #efefef "
            sty += "stop: 1 %s" % final_color
            sty += ");"
        self.lblBanner.setStyleSheet(sty)
        lwLay.addWidget(self.lblBanner)

        iconLabel = QtWidgets.QLabel()
        iconLabel.setStyleSheet("padding: 4px; background-color: %s" % final_color)
        icon = Ico.icon(Ico.FavIcon)
        iconLabel.setPixmap(icon.pixmap(QtCore.QSize( 25, 25 )))
        lwLay.addWidget(iconLabel)


        ##===============================================
        ## Central widget contains  tabBar and a stack
        centralWidget = QtWidgets.QWidget()
        centralLayout = QtWidgets.QVBoxLayout()
        centralLayout.setContentsMargins(0, 0, 0, 0)
        centralLayout.setSpacing(0)
        centralWidget.setLayout(centralLayout)
        self.setCentralWidget(centralWidget)

        tabContainer = xwidgets.hlayout()
        centralLayout.addLayout(tabContainer)

        self.tabBar = QtWidgets.QTabBar()
        self.tabBar.setMovable(False)
        self.tabBar.setTabsClosable(True)
        tabContainer.addWidget(self.tabBar)
        self.tabBar.currentChanged.connect(self.on_tab_changed)
        self.tabBar.tabCloseRequested.connect(self.on_tab_close_requested)
        tabContainer.addStretch(10)


        self.stackWidget = QtWidgets.QStackedWidget()
        centralLayout.addWidget(self.stackWidget)

        # create a progress dialog and hide
        self.progressDialog = None
        # self.progressDialog = QtWidgets.QProgressDialog(self)
        # self.progressDialog.setMinimumWidth(300)
        # self.progressDialog.setWindowIcon(Ico.icon(Ico.Busy))
        # self.progressDialog.setRange(0, 0)
        # self.progressDialog.setCancelButton(None)
        # self.progressDialog.setModal(True)
        # self.progressDialog.hide()

        #=========================================
        # Seutp basic window dims, and restore
        self.setMinimumWidth(800)
        self.setMinimumHeight(800)
        Settings.restore_window( self )
        self.load_recent()

        ## run some stuff a few moments after window shown
        QtCore.QTimer.singleShot(200, self.on_after)

    def show_progress(self):
        self.progressDialog = QtWidgets.QProgressDialog(self)
        self.progressDialog.setMinimumWidth(300)
        self.progressDialog.setWindowIcon(Ico.icon(Ico.Busy))
        self.progressDialog.setRange(0, 0)
        self.progressDialog.setCancelButton(None)
        self.progressDialog.setModal(True)
        self.progressDialog.show()

    def hide_progress(self):
        todo()

    def add_widget(self, widget, label, ico=None):

        idx = self.tabBar.addTab(label)
        self.stackWidget.addWidget(widget)
        self.tabBar.setTabIcon(idx, Ico.icon(ico))

        self.tabBar.setCurrentIndex(self.tabBar.count() - 1)
        self.stackWidget.setCurrentIndex(self.stackWidget.count() - 1)

        widget.init_load()

    def on_www_request_finished(self, xreply):

        ## we loop though all the widgets and load_reply
        ## More than one widget may be interested in data
        for i in range(0, self.stackWidget.count()):
            self.stackWidget.widget(i).load_reply(xreply)

    def closeEvent( self, event ):
        Settings.save_window( self )

    def on_quit(self):
        ret = QtWidgets.QMessageBox.warning( self, "Desktop", "Sure you want to Quit ?", QtWidgets.QMessageBox.No | QtWidgets.QMessageBox.Yes )
        if ret == QtWidgets.QMessageBox.Yes:
            G.settings.save_window(  self )
            sys.exit( 0 )

    def on_show_examples(self):
        chk = self.buttExamples.isChecked()
        self.dockExamples.setVisible(chk)

    def load_ags4_example(self, file_name):

        self.menuExamples.close()

        data, err = ags4.example(file_name)
        if err:
            pass

        proj = ogtgui_project.OGTProjectWidget()
        proj.ogtDoc.add_ags4_string(data['contents'], file_name)

        self.add_widget(proj, file_name, ico=Ico.Project)



    def on_new_project(self):

        projWidget = ogtgui_project.OGTProjectWidget()
        self.add_widget(projWidget, "New Project", ico=Ico.Project)

    def load_ags4_file(self, file_path):

        if xwidgets.file_exists_check(self, file_path) == False:
            return

        # first check its not open
        for idx in range(0, self.stackWidget.count()):
            continue
            if isinstance(self.stackWidget.widget(idx), ogtgui_project.OGTProjectWidget):

                if self.stackWidget.widget(idx).ogtDoc.source_file_path == file_path:
                    self.tabBar.setCurrentIndex(idx)
                    return

        if False:
            self.progressDialog.setWindowTitle("Loading...")
            self.progressDialog.setLabelText(file_path)
            self.progressDialog.show()

            self.add_history(file_path)

        ## ======================
        proj = ogtgui_project.OGTProjectWidget()

        self.add_widget(proj, os.path.basename(file_path), ico=Ico.Project)

        proj.add_ags4_file(file_path)

        if False:
            self.progressDialog.hide()

    def add_history(self, file_path):
        history = G.settings.get_list("history")

        if file_path in history:
            # remove if in history, so insert at top
            history.remove(file_path)
        history.insert(0, file_path)
        if len(history) > 20:
            history = history[0:20]
        Settings.save_list("history", history)
        self.load_recent()

    def load_recent(self):
        files = Settings.get_list("history")
        menu = self.actionRecent.menu()
        menu.clear()
        for f in files:
            menu.addAction(f)

    def on_open_recent(self, act):
        fn =  str(act.text())
        self.load_ags4_file(fn)

    def on_tab_changed(self, idx):
        self.stackWidget.setCurrentIndex(idx)

    def set_action_checked(self, act, state):
        act.blockSignals(True)
        act.setChecked(state)
        act.blockSignals(False)

    def on_tab_close_requested(self, idx):

        widget = self.stackWidget.widget(idx)

        if isinstance(widget, ags4_widgets.AGS4DataDictBrowser):
            self.set_action_checked(self.actionAgs4Browse, False)

        self.tabBar.removeTab(idx)
        self.stackWidget.removeWidget( widget )

    def on_browse_ags4(self):
        """Opens or switches to the :ref:`ags4_data_dict`"""

        ## Check it not already there
        for idx in range(0, self.stackWidget.count()):
            if isinstance(self.stackWidget.widget(idx), ags4_widgets.AGS4DataDictBrowser):
                ## its there so switch to
                self.stackWidget.setCurrentIndex(idx)
                self.tabBar.setCurrentIndex(idx)
                self.set_action_checked(self.actionAgs4Browse, True)
                return

        # create new instance
        browseWidget = ags4_widgets.AGS4DataDictBrowser()
        self.add_widget(browseWidget, "AGS4 Data Dict", ico=Ico.Ags4)
        ## set the menu/tbar actions checked
        self.set_action_checked(self.actionAgs4Browse, True)



    def on_open_ags_file(self):

        dial = QtWidgets.QFileDialog(self, "Select AGS File")
        dial.setFileMode(QtWidgets.QFileDialog.ExistingFile)
        #dial.setFilter(QtCore.QDir.Files)
        dial.setNameFilters(["Ags Files (*.ags *.ags4)"])

        if dial.exec_():
            fn = str(dial.selectedFiles()[0])
            self.load_ags4_file(fn)



    @staticmethod
    def show_splash():

        splashImage = QtGui.QPixmap( "../images/splash.png" )
        splashScreen = QtWidgets.QSplashScreen( splashImage )
        splashScreen.showMessage( "  Loading . . ." )
        splashScreen.show()
        return splashScreen



# -*- coding: utf-8 -*-



from Qt import QtGui, QtWidgets, QtCore, Qt, pyqtSignal

import xwidgets
from ogt.ogt_validate import OGTError

class ErrorsListModel(QtCore.QAbstractTableModel):

    class C:
        lidx = 0
        cidx = 1
        rule = 2
        err_type = 3
        description = 4
        group_code = 5
        head_code = 6

    COL_LABELS = ["Line", "Col", "Rule", " ", "Description", "Group", "Heading"]

    def __init__(self, mode, ogtDoc=None, ogtGroup=None):
        QtCore.QAbstractTableModel.__init__(self)


        self.mode = mode
        self.ogtDoc = ogtDoc
        self.ogtGroup = ogtGroup

        self.hi_cell = None

    def columnCount(self, midx=None):
        return len(self.COL_LABELS)

    def rowCount(self, midx):

        if self.mode == VIEW_ERR_MODE.document:
            return self.ogtDoc.errors_warnings_count

        if self.mode == VIEW_ERR_MODE.group:
            return self.ogtGroup.errors_warnings_count

    def get_error(self, midx):
        errors = self._get_errors()
        return errors[ midx.row() ]

    def _get_errors(self):
        if self.mode == VIEW_ERR_MODE.document:
            return self.ogtDoc.errors_list()

        elif self.mode == VIEW_ERR_MODE.group:
            return self.ogtGroup.errors_list()

    def data(self, midx, role=Qt.DisplayRole):

        ridx = midx.row()
        cidx = midx.column()

        #errors = self._get_errors()
        #err = errors[ridx]
        err = self._get_errors()[ridx]

        if role == Qt.DisplayRole:

            if cidx == self.C.err_type:
                return "E" if err.err_type == OGTError.ERR else "W"

            if cidx == self.C.description:
                #print "m-=", err.message
                return err.message if err.message else "MISS"

            if cidx == self.C.rule:
                return err.rule if err.rule else "-"

            if cidx == self.C.lidx:
                return err.cell.lidx + 1

            if cidx == self.C.cidx:
                return err.cell.cidx + 1

            if cidx == self.C.group_code:
                return err.group_code

            if cidx == self.C.head_code:
                return err.head_code
            #if col == self.C.data_count:
            #    return grp.data_rows_count()
            return "?%s/%s?" % (ridx, cidx)

        if False and role == Qt.DecorationRole:
            if cidx == self.C.group_code:
                return Ico.icon(Ico.Group)

        if False and role == Qt.FontRole:
            if cidx == self.C.group_code:
                f = QtWidgets.QFont()
                f.setBold(True)
                return f

        if role == Qt.TextAlignmentRole:
            if cidx in [self.C.description,self.C.group_code, self.C.head_code]:
                return Qt.AlignLeft
            return Qt.AlignCenter

        if role == Qt.BackgroundColorRole:

            ## highlight line and col if highlight
            if self.hi_cell and cidx in [self.C.lidx, self.C.cidx]:
                if self.hi_cell.lidx == err.cell.lidx and self.hi_cell.cidx == err.cell.cidx:
                    return QtWidgets.QColor("#EBE267")

            if cidx == self.C.err_type:
                return QtGui.QColor(err.bg)
            #return QtWidgets.QColor()


        return None # QtCore.QVariant()


    def headerData(self, idx, orient, role=None):
        if role == Qt.DisplayRole and orient == Qt.Horizontal:
            return QtCore.QVariant(self.COL_LABELS[idx])

        if role == Qt.TextAlignmentRole and orient == Qt.Horizontal:
            if idx in [self.C.description, self.C.group_code, self.C.head_code]:
                return Qt.AlignLeft
            return Qt.AlignCenter

        return QtCore.QVariant()

    def set_highlight(self, cell):

        self.hi_cell = cell
        self.reset()



class VIEW_ERR_MODE:
    document = "document"
    group = "group"
    #heading = "heading"

class ErrorsWidget( QtWidgets.QWidget ):

    sigGotoError = pyqtSignal(object)
    #sigGotoSource = pyqtSignal(int, int)
    sigErrorsFilter = pyqtSignal(bool, bool)




    def __init__( self, parent=None, mode=None, ogtGroup=None, ogtDoc=None):
        QtWidgets.QWidget.__init__( self, parent )

        self.debug = False
        self.mode = mode
        if self.mode == None:
            freak_out()

        self.mainLayout = QtWidgets.QVBoxLayout()
        self.mainLayout.setSpacing(0)
        self.mainLayout.setContentsMargins(0,0,0,0)
        self.setLayout(self.mainLayout)

        lbl = QtWidgets.QLabel()
        lbl.setText("Errors and Warnings")
        lbl.setStyleSheet("font-weight: bold; padding: 3px; background-color: #eeeeee;")
        self.mainLayout.addWidget(lbl)

        self.toolBar = QtWidgets.QToolBar()
        self.mainLayout.addWidget(self.toolBar)




        self.buttGroupFilters = QtWidgets.QButtonGroup(self)
        self.buttGroupFilters.setExclusive(False)

        self.buttWarnings = xwidgets.XToolButton(text="Show Warnings", checkable=True, checked=True)
        self.buttGroupFilters.addButton(self.buttWarnings)

        self.buttErrors = xwidgets.XToolButton(text="Show Errors", checkable=True, checked=True)
        self.buttGroupFilters.addButton(self.buttErrors)

        self.buttGroupFilters.buttonClicked.connect(self.on_update_filter)

        self.toolBar.addWidget(self.buttWarnings)
        self.toolBar.addWidget(self.buttErrors)

        #=============================
        ## Set up tree
        self.tree = QtWidgets.QTreeView()
        self.mainLayout.addWidget(self.tree, 30)

        self.model = ErrorsListModel(mode=self.mode, ogtGroup=ogtGroup, ogtDoc=ogtDoc)

        #self.proxy = ErrorsListSortFilterModel()
        #self.proxy.setSourceModel(self.model)
        self.tree.setModel(self.model)



        self.tree.setRootIsDecorated(False)
        self.tree.setExpandsOnDoubleClick(False)
        self.tree.header().setStretchLastSection(True)
        #self.tree.setSortingEnabled(True)


        #self.tree.setColumnHidden(C_ERR.err, True)
        C = ErrorsListModel.C
        #self.tree.setColumnHidden(C.search, True)
        self.tree.setColumnWidth(C.err_type, 30)
        self.tree.setColumnWidth(C.lidx, 40)
        self.tree.setColumnWidth(C.cidx, 40)
        self.tree.setColumnWidth(C.rule, 50)
        #self.tree.setColumnWidth(C.highlight, 8)

        #self.tree.itemClicked.connect(self.on_tree_item_clicked)
        self.tree.selectionModel().selectionChanged.connect(self.on_tree_selection)

    def clear(self):
        pass #print "clear"
        #self #self.tree.clear()


    def do_update(self):
        self.model.layoutChanged.emit()

    def on_tree_selection(self):

        if not self.tree.selectionModel().hasSelection():
            return

        midx = self.tree.selectionModel().selectedRows()[0]
        err = self.model.get_error(midx)
        self.sigGotoError.emit(err)



    def on_update_filter(self):
        return
        self.on_show_warnings()
        self.on_show_errors()
        self.emit_filters_sig()


    def emit_filters_sig(self):
        self.sigErrorsFilter.emit(self.buttWarnings.isChecked(), self.buttErrors.isChecked())

    def get_error_filters(self):
        return self.buttWarnings.isChecked(), self.buttErrors.isChecked()

    def select_from_cell(self, cell):

        self.model.set_highlight( cell )


    def deadselect_from_cell(self, cell):
        self.tree.blockSignals(True)

        # clear selection and  hightlight colors
        self.tree.clearSelection()
        root = self.tree.invisibleRootItem()
        for i in range(0, root.childCount()):
            root.child(i).set_bg(C_ERR.highlight, "white")

        if ridx != None and cidx != None:
            # search and hightlight row/col if any
            search = "%s-%s" % (ridx, cidx)
            items = self.tree.findItems(search, Qt.MatchExactly, C_ERR.search)
            if len(items) > 0:
                for item in items:
                    item.set_bg(C_ERR.highlight, "purple")
        self.tree.blockSignals(False)



# -*- coding: utf-8 -*-



import os

from Qt import QtGui, QtCore

from ogtgui import ICONS_PATH


class Ico(QtGui.QIcon):
    """Icons Definition and Loader

      - All icons used are listed as class constants.
      - Static methods create icons loaded from the file system

      .. todo:: try and intercept __getattr__ to return qIcon
    """

    def __init__(self, file_name=None):



        super().__init__(file_name)

    @staticmethod
    def get_icon_path(file_name):
        """
        Args:
            file_name (str): Icon filename
        Returns:
            str: Icons full path
        """
        return os.path.join(ICONS_PATH, file_name)

    @staticmethod
    def icon( file_name, pixmap=False, size=None):
        """Return a `QIcon`

        Args:
            file_name (str): Icon filename to load
            pixmap (bool): Return a pixmap instead of QIcon object
            size (int): Default size
        Returns:
            QIcon: new icon or pixmap
        """
        #print(Ico.get_icon_path(file_name))
        if file_name == None:
            qicon = QtGui.QIcon()
        else:
            qicon = QtGui.QIcon( Ico.get_icon_path(file_name) )

        if pixmap:
            return qicon.pixmap( QtCore.QSize( 16, 16 ) )

        if size:
            pass #icon2 = QtGui.QIcon( icon.pixmap( QtCore.QSize(size, size) ) )
            #return icon2

        return qicon

    Add = "add.png"

    Ags3 = "ags3.svg.png"
    Ags4 = "ags4.svg.png"

    AgsAbbrev = "font.png"
    AgsAbbrevItem = "textfield_rename.png"

    AgsField = "textfield.png"

    AgsGroup = "layout.png"
    AgsGroups = "layers.png"
    AgsHeading = "layout_header.png"
    AgsNotes = "note.png"

    Busy = "arrow_in.png"
    BulletDown = "bullet_arrow_down.png"
    BulletBlack = "bullet_black.png"

    Cancel = "bullet_black.png"
    Clear = "control.png"


    Document = "blue-folder-open-document.png"

    Export = "fill.png"
    Excel = "document-excel.png"

    FavIcon = "logo.png"

    FilterOn = "bullet_yellow.png"
    FilterOff = "bullet_blue.png"

    Folder = 'folder.png'
    FolderAdd = 'folder_add.png'
    FolderCopy = 'folder_page.png'
    FolderDelete = 'folder_delete.png'
    FolderOpen = 'folder_go.png'
    FolderRename = 'folder_edit.png'

    Groups = "tables-stacks.png"
    Group = "table.png"

    KeyField = "key.png"

    Import = "arrow-join-270.png"

    Project = "blue-document-list.png"
    Plus = "plus-small-white.png"

    Map = "map.png"
    Minus = "minus-small.png"

    Quit = 'control_eject.png'

    Refresh = "arrow_refresh.png"

    Save = "tick.png"
    Samples = "tag_yellow.png"
    Schedule = "table-heatmap.png"
    Source = "document-text.png"
    Summary = "page_white_swoosh.png"

    TestPoint = "bullet_purple.png"

    TypeCheckBox = "ui-checkbox.png"
    TypeDate = "ui-combo-box-calendar.png"
    TypeDecimal = "ui-text-field-password-green.png"
    TypeID = "type_id.svg.png"
    TypePicklist = "ui-combo-box-blue.png"
    TypeSci = "ui-text-field-password-yellow.png"
    TypeStandard = "ui-.png"
    TypeText = "ui-text-field.png"
    TypeUnknown = "ui-text-field-hidden.png"


    TickOn = "tick.png"
    TickOff = "bullet_black.png"

    Zip = "compress.png"




"""This module provides the Data Dict (:class:`DD` module variable), an instance of :class:`AGS4_DataDict`

- This is used to present and validate :term:`ags4` files.
- See :ref:`ags4_data_dict`

"""

import os

from ogt import PROJECT_ROOT_PATH
from ogt import utils
from ogt.ogt_validate import OGTError

DICT_DTYP = "DICT_DTYP"


def ags4dd_file_path():
    """
    :rtype: str
    :return: Absolute path to the `ags4.min.json` data dict file
    """
    return os.path.abspath(os.path.join(PROJECT_ROOT_PATH, "ogtstatic", "ags4.min.json"))


class DESCRIPTOR:
    """Constants defining the data descriptors (See :ref:`ags4_rule_3`)

       - The data descriptor is in the first column of every AGS row
       - If the first column is not one of above, then its an error.
    """
    group = "GROUP"
    heading = "HEADING"
    unit = "UNIT"
    data_type = "TYPE"
    data = "DATA"


class AGS4DataDict:
    """This class contains all the ags4 data dict """

    @staticmethod
    def group_header():
        """The list of headers as required and defined in :ref:`ags4_rule_2b`

        :type: list[str]
        :return: A list of `descriptors` in order
        """
        return [
            DESCRIPTOR.group,
            DESCRIPTOR.heading,
            DESCRIPTOR.unit,
            DESCRIPTOR.data_type
        ]

    @staticmethod
    def descriptors():
        """
        :return: A list of descriptors in correct order (see :ref:`ags4_rule_3`)
        :rtype: list
        """
        return [
            DESCRIPTOR.group, DESCRIPTOR.heading, DESCRIPTOR.unit, DESCRIPTOR.data_type, DESCRIPTOR.data
        ]

    def __init__(self):

        self.initialised = False
        """True if loaded with no errors (see :meth:`AGS4_DataDict.setup_initialise`"""

        self._data = None
        self._groups_dict = None
        self._groups_index = None
        self.abbrs_dict = None
        self._data_types = None
        self._units = None

        self._types_lookup_cache = None

        self.setup_initialise()


    def setup_initialise(self):
        """Check env is sane and loads the ags data dict file"""

        if not os.path.exists(ags4dd_file_path()):
            return "Missing ags4 data dict %s" % ags4dd_file_path()

        data, err = utils.read_json_file(ags4dd_file_path())
        if err:
            print("error..", err)
            return err

        self.abbrs_dict = data['abbrs']

        self._groups_dict = data['groups']

        self._groups_index = sorted(self._groups_dict.keys())

        # Walk groups/headings and create a _words dict for lookup
        self.words = {}
        # TODO
        # if False:
        #     for grp in self._groups_dict.itervalues():
        #         self.words[grp['group_code']] = dict(type="group",
        #                                               code=grp['group_code'],
        #                                               description=grp['group_description'])
        #         for head in grp['headings']:
        #             if not head['head_code'] in self.words: # not add dupes
        #                 self.words[head['head_code']] = dict(type="heading",
        #                                                       code=head['head_code'],
        #                                                       description=head['head_description'])

        self.initialised = True

    def groups_count(self):
        """
        :rtype: int
        :return:  Count of groups
        """
        return len(self._groups_index)

    def group_by_index(self, idx):
        """
        :param idx: index
        :type idx: int
        :rtype: dict
        :return:  dict at given index
        """
        return self._groups_dict[self._groups_index[idx]]

    def groups_list(self):
        return [self._groups_dict[gki] for gki in self._groups_index]

    def groups_dict(self):
        """
        :rtype: dict[group_code] -> dict
        :return: All groups with group code as key
        """
        return self._groups_dict

    def group(self, group_code):
        """
        :param group_code: four character group code
        :type group_code: str
        :rtype: dict or None
        :return:  the group data if successful, else `None`
        """
        return self._groups_dict.get(group_code)

    @property
    def data_types(self):
        """list of :term:`TYPE` s  (:ref:`ags4_rule_17`)

        :type: list[dict]
        """
        return self.abbrs_dict.get("DICT_DTYP")

    @property
    def data_types_list(self):
        """List of TYPE codes

        :type: list[str]
        """
        return [rec['code'] for rec in self.data_types]
        return self.abbrs_dict.get("DICT_DTYP")

    @property
    def data_types_dict(self):
        """data types the abbr code as key

        :type: dict[code] => dict
        """
        dic = {}
        for rec in self.data_types:
            dic[rec['code']] = rec
        return dic

    @property
    def units_list(self):
        """list of units (:ref:`ags4_rule_8`)

        :type: list
        """
        return self.abbrs_dict.get("UNIT_UNIT")

    def abbrs(self, head_code):
        """Abbreviations for heading (:ref:`ags4_rule_16`)

        :param head_code: heading code
        :type head_code: str
        :rtype: list[dict] else None
        :return:  a dict with the picklist if successful else `None`
        """
        recs = self.abbrs_dict.get(head_code)
        if recs:
            return recs
        return None

    def headings(self, group_code):
        """
        :param group_code: four characteter group code
        :type group_code: str
        :rtype: list[dict]
        :return: Headings in the group
        """
        if group_code not in self._groups_dict:
            return None
        return self._groups_dict[group_code]["headings"]

    def headings_index(self, group_code):
        """
        :param group_code: Four char group code
        :type group_code: str
        :rtype: list[str] or None
        :return: list of heading codes if group exist
        """
        if group_code not in self.groups_list():
            return None
        return [h['head_code'] for h in self._groups_dict[group_code]["headings"]]

    def heading(self, head_code):
        """
        Attempts to get the heading data from head_code (this
        may not work as expected as some groups are missing eg `SPEC`)

        :param hcode: heading code
        :type hcode: str
        :rtype: (dict or None, str)
        :return: Tuple if  data and error
        """
        gc = head_code.split("_")[0]
        if gc not in self._groups_dict:
            return None, "Invalid Group `%s` in head code  %s " % (gc, head_code)

        if head_code in self._groups_dict[gc]["headings"]:
            return self._groups_dict[gc]["headings"][head_code], None
        return None, "Invalid headcode `%s` in group  %s " % (head_code, gc)

    def classified_groups(self):
        """
        :rtype: dict[classificaton] => list[group_dict]
        :return: groups nested by classification
        """
        classes = {}
        for gcode, grp in self.groups_dict().items():
            cls = grp['class']
            if cls not in classes:
                classes[cls] = {}
            classes[cls][gcode] = grp
        return classes


DD = AGS4DataDict()
"""Global Instance of :class:`AGS4DataDict` initialised on import """


def summary_to_string(summary):

    ret = ""

    for sum_dic in summary:
        ret += report_to_string(sum_dic)

    return ret


def report_to_string(report):
    ret = "=" * 50
    ret += "\n%s\n" % report['file_path']
    ret += "-" * 50
    ret += "\n"

    for rule, dic in sorted(report['rules'].items()):
        if len(dic['errors']) == 0 and len(dic['problems']) == 0:
            pass
        else:
            ret += "%s:" % rule.replace("_", " ").capitalize()
            ret += " FAIL\n"

            if len(dic['errors']):
                ret += "\tSystem Errors:\n"
                for item in dic['errors']:
                    ret += "\t\t%s\n" % item

            lp = len(dic['problems'])
            if lp:
                ret += "\tProblem: %s %s\n" % (lp, "item" if lp == 1 else "items")
                for prob in dic['problems']:
                    ret += "\t\t%s\n" % prob

    return ret


def rule_1(doc):
    """ Validate :ref:`ags4_rule_1`

    :param raw_str:
    :rtype: tuple
    :return:
        - A list of ags_errors
        - A list of sys_errors
    """
    warnings = []
    errors = []
    try:
        # try and force to ascii in a one liner
        doc.source.decode("ascii")
        return warnings, errors

    except UnicodeDecodeError:

        # raw_str has problem somewhere.
        # so split to lines, = line
        # and then check every char in line = column
        # if char fail, we add to errrors
        errors.append("Unicode decode error")
        for lidx, line in enumerate(doc.source.split("\n")):

            for cidx, charlie in enumerate(line):
                # TODO . check visible.. maybe some < 30 chard are a hacker
                if ord(charlie) < 128:
                    pass
                else:
                    warnings.append(dict(line=lidx+1, column=cidx+1, illegal=[]))
    return warnings, errors


def rule_2(doc):
    """ Validate :ref:`ags4_rule_2`

    :param doc:
    :type doc: :class:`~ogt.ogt_doc.OGTDocument`
    :rtype: tuple
    :return:
        - a list of ags_errors
        - a list of sys errors
    """
    problems = []
    # report = []
    errors = []

    RULE = 2

    # check more than one group
    if doc.groups_count() < 2:
        p = OGTError(message="Need more than one group")
        problems.append(p)

    # Each data GROUP shall comprise a number of GROUP HEADER
    # TODO

    # rows must have one or more ref DATA rows
    for group_code, grp in doc.groups.items():
        if len(grp.data) == 0:
            p = OGTError(group_code, rule=RULE)
            problems.append(p)
            # report.append("group `%s` has no DATA rows" % group_code)

    # Rule 2b - CRLF end each line
    # so we in unix.. so split source with \n and check there's a \r at end
    # can only be done with raw files
    # TODO
    l_rep = []

    # Rule 2c GROUP HEADER
    # loop each group, and check the headers
    l_rep = []
    for group_code, grp in doc.groups.items():

        for idx, dd in enumerate(DD.group_header()):

            if grp.csv_rows()[0] != dd:
                p = OGTError(group_code, rule=RULE)

                p.message = "GROUP_HEADER error. "
                p.message += "Incorrect order line "
                problems.append(p)

    if len(l_rep):
        # report.append(l_rep)
        pass

    return problems, errors


EXAMPLES_FILE = os.path.join(PROJECT_ROOT_PATH, "ogtstatic", "ags4_examples.min.json")


def examples_all():
    # TODO check dir exists

    data, err = utils.read_json_file(EXAMPLES_FILE)
    return data, err


def examples_list():
    data, err = utils.read_json_file(EXAMPLES_FILE)
    if err:
        return None, err
    return [{"file_name": r['file_name']} for r in data['ags4_examples']], None


def example(file_name):

    data, err = examples_all()
    if err:
        return None, err
    for r in data['ags4_examples']:
        if r['file_name'] == file_name:
            return r, None
    return None, "Example `%s` not found " % file_name

